@@ -10,5 +10,10 @@ int main(void)
 	};
 
 	syscall(SYS_perf_event_open, &attr, 0, -1, -1, 0);
+
+	if (PERF_RECORD_SWITCH != 14 ||
+	    PERF_RECORD_MISC_SWITCH_OUT != (1 << 13))
+		return -1;
+
 	return 0;
 }
