@@ -30,11 +30,6 @@
 #include "utils/filter.h"
 #include "utils/script.h"
 
-/* could be defined in mcount-arch.h */
-#ifndef  ARCH_SUPPORT_AUTO_RECOVER
-# define ARCH_SUPPORT_AUTO_RECOVER  0
-#endif
-
 /* time filter in nsec */
 uint64_t mcount_threshold;
 
@@ -85,19 +80,40 @@ static enum filter_mode __maybe_unused mcount_filter_mode = FILTER_MODE_NONE;
 /* tree of trigger actions */
 static struct rb_root __maybe_unused mcount_triggers = RB_ROOT;
 
-/* number of active thread running mcount code */
-static int mcount_active;
+/* bitmask of active watch points */
+static unsigned long __maybe_unused mcount_watchpoints;
+
+/* whether caller filter is activated */
+static bool __maybe_unused mcount_has_caller;
+
+/* address of function will be called when a function returns */
+unsigned long mcount_return_fn;
+
+/* disassembly engine for dynamic code patch */
+static struct mcount_disasm_engine disasm;
+
+__weak void dynamic_return(void) { }
 
 #ifdef DISABLE_MCOUNT_FILTER
 
 static void mcount_filter_init(enum uftrace_pattern_type ptype, char *dirname,
 			       bool force)
 {
+	if (getenv("UFTRACE_SRCLINE") == NULL)
+		return;
+
+	symtabs.maps->mod = load_module_symtab(&symtabs, symtabs.maps->libname);
+
 	/* use debug info if available */
 	prepare_debug_info(&symtabs, ptype, NULL, NULL, false, force);
 	save_debug_info(&symtabs, dirname);
 }
 
+static void mcount_filter_finish(void)
+{
+	finish_debug_info(&symtabs);
+}
+
 #else
 
 static void prepare_pmu_trigger(struct rb_root *root)
@@ -124,6 +140,260 @@ static void prepare_pmu_trigger(struct rb_root *root)
 	}
 }
 
+/* be careful: this can be called from signal handler */
+static void mcount_finish_trigger(void)
+{
+	if (mcount_global_flags & MCOUNT_GFL_FINISH)
+		return;
+
+	/* mark other threads can see the finish flag */
+	mcount_global_flags |= MCOUNT_GFL_FINISH;
+}
+
+static LIST_HEAD(siglist);
+
+struct signal_trigger_item {
+	struct list_head list;
+	int sig;
+	struct uftrace_trigger tr;
+};
+
+static struct uftrace_trigger * get_signal_trigger(int sig)
+{
+	struct signal_trigger_item *item;
+
+	list_for_each_entry(item, &siglist, list) {
+		if (item->sig == sig)
+			return &item->tr;
+	}
+
+	return NULL;
+}
+
+static void add_signal_trigger(int sig, const char *name,
+			       struct uftrace_trigger *tr)
+{
+	struct signal_trigger_item *item;
+
+	item = xmalloc(sizeof(*item));
+	item->sig = sig;
+	memcpy(&item->tr, tr, sizeof(*tr));
+
+	pr_dbg("add signal trigger: %s (%d), flags = %lx\n",
+	       name, sig, (unsigned long)tr->flags);
+
+	list_add(&item->list, &siglist);
+}
+
+static void mcount_signal_trigger(int sig)
+{
+	struct uftrace_trigger *tr;
+
+	tr = get_signal_trigger(sig);
+	if (tr == NULL)
+		return;
+
+	pr_dbg("got signal %d\n", sig);
+
+	if (tr->flags & TRIGGER_FL_TRACE_ON) {
+		mcount_enabled = true;
+	}
+	if (tr->flags & TRIGGER_FL_TRACE_OFF) {
+		mcount_enabled = false;
+	}
+	if (tr->flags & TRIGGER_FL_FINISH) {
+		mcount_finish_trigger();
+	}
+}
+
+#define SIGTABLE_ENTRY(s)  { #s, s }
+
+static const struct sigtable {
+	const char *name;
+	int sig;
+} sigtable[] = {
+	SIGTABLE_ENTRY(SIGHUP),
+	SIGTABLE_ENTRY(SIGINT),
+	SIGTABLE_ENTRY(SIGQUIT),
+	SIGTABLE_ENTRY(SIGILL),
+	SIGTABLE_ENTRY(SIGTRAP),
+	SIGTABLE_ENTRY(SIGABRT),
+	SIGTABLE_ENTRY(SIGBUS),
+	SIGTABLE_ENTRY(SIGFPE),
+	SIGTABLE_ENTRY(SIGKILL),
+	SIGTABLE_ENTRY(SIGUSR1),
+	SIGTABLE_ENTRY(SIGSEGV),
+	SIGTABLE_ENTRY(SIGUSR2),
+	SIGTABLE_ENTRY(SIGPIPE),
+	SIGTABLE_ENTRY(SIGALRM),
+	SIGTABLE_ENTRY(SIGTERM),
+	SIGTABLE_ENTRY(SIGSTKFLT),
+	SIGTABLE_ENTRY(SIGCHLD),
+	SIGTABLE_ENTRY(SIGCONT),
+	SIGTABLE_ENTRY(SIGSTOP),
+	SIGTABLE_ENTRY(SIGTSTP),
+	SIGTABLE_ENTRY(SIGTTIN),
+	SIGTABLE_ENTRY(SIGTTOU),
+	SIGTABLE_ENTRY(SIGURG),
+	SIGTABLE_ENTRY(SIGXCPU),
+	SIGTABLE_ENTRY(SIGXFSZ),
+	SIGTABLE_ENTRY(SIGVTALRM),
+	SIGTABLE_ENTRY(SIGPROF),
+	SIGTABLE_ENTRY(SIGWINCH),
+	SIGTABLE_ENTRY(SIGIO),
+	SIGTABLE_ENTRY(SIGPWR),
+	SIGTABLE_ENTRY(SIGSYS),
+};
+
+#undef SIGTABLE_ENTRY
+
+static int parse_sigspec(char *spec, struct uftrace_filter_setting *setting)
+{
+	char *pos, *tmp;
+	unsigned i;
+	int sig = -1;
+	int off = 0;
+	const char *signame = NULL;
+	bool num_spec = false;
+	char num_spec_str[16];
+	struct uftrace_trigger tr = {
+		.flags = 0,
+	};
+	struct sigaction old_sa;
+	struct sigaction sa = {
+		.sa_handler  = mcount_signal_trigger,
+		.sa_flags    = SA_RESTART,
+	};
+	const char *sigrtm   = "SIGRTM";
+	const char *sigrtmin = "SIGRTMIN";
+	const char *sigrtmax = "SIGRTMAX";
+
+	pos = strchr(spec, '@');
+	if (pos == NULL)
+		return -1;
+	*pos = '\0';
+
+	if (isdigit(spec[0]))
+		num_spec = true;
+	else if (strncmp(spec, "SIG", 3))
+		off = 3;  /* skip "SIG" prefix */
+
+	for (i = 0; i < ARRAY_SIZE(sigtable); i++) {
+		if (num_spec) {
+			int num = strtol(spec, &tmp, 0);
+
+			if (num == sigtable[i].sig) {
+				sig = num;
+				signame = sigtable[i].name;
+				break;
+			}
+
+			continue;
+		}
+
+		if (!strcmp(sigtable[i].name + off, spec)) {
+			sig = sigtable[i].sig;
+			signame = sigtable[i].name;
+			break;
+		}
+	}
+
+	/* real-time signals */
+	if (!strncmp(spec, sigrtm + off, 6 - off)) {
+		if (!strncmp(spec, sigrtmin + off, 8 - off))
+			sig = SIGRTMIN + strtol(&spec[8 - off], NULL, 0);
+		if (!strncmp(spec, sigrtmax + off, 8 - off))
+			sig = SIGRTMAX + strtol(&spec[8 - off], NULL, 0);
+		signame = spec;
+	}
+
+	if (sig == -1 && num_spec) {
+		int sigrtmid = (SIGRTMIN + SIGRTMAX) / 2;
+
+		sig = strtol(spec, &tmp, 0);
+
+		/* SIGRTMIN/MAX might not be constant, avoid switch/case */
+		if (sig == SIGRTMIN) {
+			strcpy(num_spec_str, "SIGRTMIN");
+		}
+		else if (SIGRTMIN < sig && sig <= sigrtmid) {
+			snprintf(num_spec_str, sizeof(num_spec_str), "%s+%d",
+				 "SIGRTMIN", sig - SIGRTMIN);
+		}
+		else if (sigrtmid < sig && sig < SIGRTMAX) {
+			snprintf(num_spec_str, sizeof(num_spec_str), "%s-%d",
+				 "SIGRTMAX", SIGRTMAX - sig);
+		}
+		else if (sig == SIGRTMAX) {
+			strcpy(num_spec_str, "SIGRTMAX");
+		}
+		else {
+			sig = -1;
+		}
+		signame = num_spec_str;
+	}
+
+	if (sig == -1) {
+		pr_use("failed to parse signal: %s\n", spec);
+		return -1;
+	}
+
+	/* setup_trigger_action() requires the '@' sign */
+	*pos = '@';
+
+	tmp = NULL;
+	if (setup_trigger_action(spec, &tr, &tmp, TRIGGER_FL_SIGNAL, setting) < 0)
+		return -1;
+
+	if (tmp != NULL) {
+		pr_warn("invalid signal action: %s\n", tmp);
+		free(tmp);
+		return -1;
+	}
+
+	add_signal_trigger(sig, signame, &tr);
+	if (sigaction(sig, &sa, &old_sa) < 0) {
+		pr_warn("cannot overwrite signal handler for %s\n", spec);
+		sigaction(sig, &old_sa, NULL);
+		return -1;
+	}
+
+	return 0;
+}
+
+static int mcount_signal_init(char *sigspec,
+			      struct uftrace_filter_setting *setting)
+{
+	struct strv strv = STRV_INIT;
+	char *spec;
+	int i;
+	int ret = 0;
+
+	if (sigspec == NULL)
+		return 0;
+
+	strv_split(&strv, sigspec, ";");
+
+	strv_for_each(&strv, spec, i) {
+		if (parse_sigspec(spec, setting) < 0)
+			ret = -1;
+	}
+	strv_free(&strv);
+
+	return ret;
+}
+
+static void mcount_signal_finish(void)
+{
+	struct signal_trigger_item *item;
+
+	while (!list_empty(&siglist)) {
+		item = list_first_entry(&siglist, typeof(*item), list);
+		list_del(&item->list);
+		free(item);
+	}
+}
+
 static void mcount_filter_init(enum uftrace_pattern_type ptype, char *dirname,
 			       bool force)
 {
@@ -132,45 +402,70 @@ static void mcount_filter_init(enum uftrace_pattern_type ptype, char *dirname,
 	char *argument_str  = getenv("UFTRACE_ARGUMENT");
 	char *retval_str    = getenv("UFTRACE_RETVAL");
 	char *autoargs_str  = getenv("UFTRACE_AUTO_ARGS");
+	char *caller_str    = getenv("UFTRACE_CALLER");
+
+	struct uftrace_filter_setting filter_setting = {
+		.ptype		= ptype,
+		.auto_args	= false,
+		.allow_kernel	= false,
+		.lp64		= host_is_lp64(),
+		.arch		= host_cpu_arch(),
+	};
+	bool needs_debug_info = false;
 
 	load_module_symtabs(&symtabs);
 
+	mcount_signal_init(getenv("UFTRACE_SIGNAL"), &filter_setting);
+
 	/* setup auto-args only if argument/return value is used */
 	if (argument_str || retval_str || autoargs_str ||
 	    (trigger_str && (strstr(trigger_str, "arg") ||
 			     strstr(trigger_str, "retval")))) {
-		setup_auto_args();
+		setup_auto_args(&filter_setting);
+		needs_debug_info = true;
 	}
 
+	if (getenv("UFTRACE_SRCLINE"))
+		needs_debug_info = true;
+
 	/* use debug info if available */
-	prepare_debug_info(&symtabs, ptype, argument_str, retval_str,
-			   !!autoargs_str, force);
-	save_debug_info(&symtabs, dirname);
+	if (needs_debug_info) {
+		prepare_debug_info(&symtabs, ptype, argument_str, retval_str,
+				   !!autoargs_str, force);
+		save_debug_info(&symtabs, dirname);
+	}
 
 	uftrace_setup_filter(filter_str, &symtabs, &mcount_triggers,
-			     &mcount_filter_mode, false, ptype);
+			     &mcount_filter_mode, &filter_setting);
 	uftrace_setup_trigger(trigger_str, &symtabs, &mcount_triggers,
-			      &mcount_filter_mode, false, ptype);
+			      &mcount_filter_mode, &filter_setting);
 	uftrace_setup_argument(argument_str, &symtabs, &mcount_triggers,
-			       false, ptype);
+			       &filter_setting);
 	uftrace_setup_retval(retval_str, &symtabs, &mcount_triggers,
-			     false, ptype);
+			     &filter_setting);
+
+	if (caller_str) {
+		uftrace_setup_caller_filter(caller_str, &symtabs,
+					    &mcount_triggers, &filter_setting);
+
+		if (uftrace_count_filter(&mcount_triggers,
+					 TRIGGER_FL_CALLER) != 0)
+			mcount_has_caller = true;
+	}
 
 	if (autoargs_str) {
-		char *autoarg = get_auto_argspec_str();
-		char *autoret = get_auto_retspec_str();
+		char *autoarg = ".";
+		char *autoret = ".";
 
-		if (debug_info_has_argspec(&symtabs.dinfo)) {
-			if (ptype == PATT_REGEX)
-				autoarg = autoret = ".";
-			else  /* PATT_GLOB */
-				autoarg = autoret = "*";
-		}
+		if (ptype == PATT_GLOB)
+			autoarg = autoret = "*";
+
+		filter_setting.auto_args = true;
 
 		uftrace_setup_argument(autoarg, &symtabs,
-				       &mcount_triggers, true, ptype);
+				       &mcount_triggers, &filter_setting);
 		uftrace_setup_retval(autoret, &symtabs,
-				     &mcount_triggers, true, ptype);
+				     &mcount_triggers, &filter_setting);
 	}
 
 	if (getenv("UFTRACE_DEPTH"))
@@ -196,6 +491,45 @@ static void mcount_filter_release(struct mcount_thread_data *mtdp)
 	mtdp->argbuf = NULL;
 }
 
+static void mcount_filter_finish(void)
+{
+	uftrace_cleanup_filter(&mcount_triggers);
+	finish_auto_args();
+
+	finish_debug_info(&symtabs);
+
+	finish_pmu_event();
+	mcount_signal_finish();
+}
+
+static void mcount_watch_init(void)
+{
+	char *watch_str   = getenv("UFTRACE_WATCH");
+	struct strv watch = STRV_INIT;
+	char *str;
+	int i;
+
+	if (watch_str == NULL)
+		return;
+
+	strv_split(&watch, watch_str, ";");
+
+	strv_for_each(&watch, str, i) {
+		if (!strcasecmp(str, "cpu"))
+			mcount_watchpoints = MCOUNT_WATCH_CPU;
+	}
+	strv_free(&watch);
+}
+
+static void mcount_watch_setup(struct mcount_thread_data *mtdp)
+{
+	mtdp->watch.cpu = -1;
+}
+
+static void mcount_watch_release(struct mcount_thread_data *mtdp)
+{
+}
+
 #endif /* DISABLE_MCOUNT_FILTER */
 
 static void send_session_msg(struct mcount_thread_data *mtdp, const char *sess_id)
@@ -231,24 +565,61 @@ static void send_session_msg(struct mcount_thread_data *mtdp, const char *sess_i
 	}
 }
 
+static void mcount_trace_finish(bool send_msg)
+{
+	static pthread_mutex_t finish_lock = PTHREAD_MUTEX_INITIALIZER;
+	static bool trace_finished = false;
+
+	pthread_mutex_lock(&finish_lock);
+	if (trace_finished)
+		goto unlock;
+
+	/* dtor for script support */
+	if (SCRIPT_ENABLED && script_str)
+		script_uftrace_end();
+
+	/* notify to uftrace that we're finished */
+	if (send_msg)
+		uftrace_send_message(UFTRACE_MSG_FINISH, NULL, 0);
+
+	if (pfd != -1) {
+		close(pfd);
+		pfd = -1;
+	}
+
+	trace_finished = true;
+	pr_dbg("mcount trace finished\n");
+
+unlock:
+	pthread_mutex_unlock(&finish_lock);
+}
+
 /* to be used by pthread_create_key() */
-static void mtd_dtor(void *arg)
+void mtd_dtor(void *arg)
 {
 	struct mcount_thread_data *mtdp = arg;
 	struct uftrace_msg_task tmsg;
 
-	if (mtdp->rstack == NULL)
+	if (mtdp->dead)
 		return;
 
+	if (mcount_should_stop())
+		mcount_trace_finish(true);
+
 	/* this thread is done, do not enter anymore */
 	mtdp->recursion_marker = true;
+	mtdp->dead = true;
 
 	mcount_rstack_restore(mtdp);
 
-	free(mtdp->rstack);
-	mtdp->rstack = NULL;
+	if (ARCH_CAN_RESTORE_PLTHOOK || !mcount_rstack_has_plthook(mtdp)) {
+		free(mtdp->rstack);
+		mtdp->rstack = NULL;
+		mtdp->idx = 0;
+	}
 
 	mcount_filter_release(mtdp);
+	mcount_watch_release(mtdp);
 	finish_mem_region(&mtdp->mem_regions);
 	shmem_finish(mtdp);
 
@@ -259,36 +630,23 @@ static void mtd_dtor(void *arg)
 	uftrace_send_message(UFTRACE_MSG_TASK_END, &tmsg, sizeof(tmsg));
 }
 
-static void mcount_trace_finish(void)
+void __mcount_guard_recursion(struct mcount_thread_data *mtdp)
 {
-	static volatile bool trace_finished;
-
-	if (trace_finished)
-		return;
-
-	/* dtor for script support */
-	if (SCRIPT_ENABLED && script_str)
-		script_uftrace_end();
-
-	if (pfd != -1) {
-		close(pfd);
-		pfd = -1;
-	}
-
-	trace_finished = true;
+	mtdp->recursion_marker = true;
 }
 
-bool mcount_guard_recursion(struct mcount_thread_data *mtdp, bool force)
+void __mcount_unguard_recursion(struct mcount_thread_data *mtdp)
+{
+	mtdp->recursion_marker = false;
+}
+
+bool mcount_guard_recursion(struct mcount_thread_data *mtdp)
 {
 	if (unlikely(mtdp->recursion_marker))
 		return false;
 
-	__sync_add_and_fetch(&mcount_active, 1);
-
-	if (!force && unlikely(mcount_should_stop())) {
+	if (unlikely(mcount_should_stop())) {
 		mtd_dtor(mtdp);
-		if (__sync_sub_and_fetch(&mcount_active, 1) == 0)
-			mcount_trace_finish();
 		return false;
 	}
 
@@ -298,17 +656,10 @@ bool mcount_guard_recursion(struct mcount_thread_data *mtdp, bool force)
 
 void mcount_unguard_recursion(struct mcount_thread_data *mtdp)
 {
-	int nr_active;
-
 	mtdp->recursion_marker = false;
 
-	nr_active = __sync_sub_and_fetch(&mcount_active, 1);
-
-	if (mcount_should_stop()) {
+	if (unlikely(mcount_should_stop()))
 		mtd_dtor(mtdp);
-		if (nr_active == 0)
-			mcount_trace_finish();
-	}
 }
 
 static struct sigaction old_sigact[2];
@@ -340,6 +691,9 @@ static void segv_handler(int sig, siginfo_t *si, void *ctx)
 	if (check_thread_data(mtdp))
 		goto out;
 
+	if (mtdp->idx <= 0)
+		goto out;
+
 	mcount_rstack_restore(mtdp);
 
 	idx = mtdp->idx - 1;
@@ -348,17 +702,19 @@ static void segv_handler(int sig, siginfo_t *si, void *ctx)
 	record_trace_data(mtdp, rstack, NULL);
 
 	if (dbg_domain[PR_DOMAIN]) {
-		for (idx = 0; idx < (int)ARRAY_SIZE(sigsegv_codes); idx++) {
+		int i;
+
+		for (i = 0; i < (int)ARRAY_SIZE(sigsegv_codes); i++) {
 			if (sig != SIGSEGV)
 				break;
 
-			if (si->si_code == sigsegv_codes[idx].code) {
+			if (si->si_code == sigsegv_codes[i].code) {
 				pr_red("Segmentation fault: %s (addr: %p)\n",
-				       sigsegv_codes[idx].msg, si->si_addr);
+				       sigsegv_codes[i].msg, si->si_addr);
 				break;
 			}
 		}
-		if (sig != SIGSEGV || idx == (int)ARRAY_SIZE(sigsegv_codes)) {
+		if (sig != SIGSEGV || i == (int)ARRAY_SIZE(sigsegv_codes)) {
 			pr_red("process crashed by signal %d: %s (si_code: %d)\n",
 			       sig, strsignal(sig), si->si_code);
 		}
@@ -412,18 +768,22 @@ struct mcount_thread_data * mcount_prepare(void)
 	struct mcount_thread_data *mtdp = &mtd;
 	struct uftrace_msg_task tmsg;
 
+	if (unlikely(mcount_should_stop()))
+		return NULL;
+
 	/*
 	 * If an executable implements its own malloc(),
 	 * following recursion could occur
 	 *
 	 * mcount_entry -> mcount_prepare -> xmalloc -> mcount_entry -> ...
 	 */
-	if (!mcount_guard_recursion(mtdp, false))
+	if (!mcount_guard_recursion(mtdp))
 		return NULL;
 
 	compiler_barrier();
 
 	mcount_filter_setup(mtdp);
+	mcount_watch_setup(mtdp);
 	mtdp->rstack = xmalloc(mcount_rstack_max * sizeof(*mtd.rstack));
 
 	pthread_once(&once_control, mcount_init_file);
@@ -445,19 +805,10 @@ struct mcount_thread_data * mcount_prepare(void)
 
 static void mcount_finish(void)
 {
-	if (mcount_global_flags & MCOUNT_GFL_FINISH)
-		return;
+	if (!mcount_should_stop())
+		mcount_trace_finish(false);
 
 	mcount_global_flags |= MCOUNT_GFL_FINISH;
-
-	uftrace_send_message(UFTRACE_MSG_FINISH, NULL, 0);
-
-	__sync_synchronize();
-
-	if (mcount_active == 0) {
-		mtd_dtor(&mtd);
-		mcount_trace_finish();
-	}
 }
 
 static bool mcount_check_rstack(struct mcount_thread_data *mtdp)
@@ -497,9 +848,9 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp,
 
 	uftrace_match_filter(child, &mcount_triggers, tr);
 
-	pr_dbg3(" tr->flags: %lx, filter mode, count: [%d] %d/%d\n",
-		tr->flags, mcount_filter_mode, mtdp->filter.in_count,
-		mtdp->filter.out_count);
+	pr_dbg3(" tr->flags: %x, filter mode: %d, count: %d/%d, depth: %d\n",
+		tr->flags, tr->fmode, mtdp->filter.in_count,
+		mtdp->filter.out_count, mtdp->filter.depth);
 
 	if (tr->flags & TRIGGER_FL_FILTER) {
 		if (tr->fmode == FILTER_MODE_IN)
@@ -634,7 +985,8 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 	rstack->filter_time  = mtdp->filter.saved_time;
 
 #define FLAGS_TO_CHECK  (TRIGGER_FL_FILTER | TRIGGER_FL_RETVAL |	\
-			 TRIGGER_FL_TRACE | TRIGGER_FL_FINISH)
+			 TRIGGER_FL_TRACE | TRIGGER_FL_FINISH |		\
+			 TRIGGER_FL_CALLER)
 
 	if (tr->flags & FLAGS_TO_CHECK) {
 		if (tr->flags & TRIGGER_FL_FILTER) {
@@ -653,9 +1005,12 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 		if (tr->flags & TRIGGER_FL_TRACE)
 			rstack->flags |= MCOUNT_FL_TRACE;
 
+		if (tr->flags & TRIGGER_FL_CALLER)
+			rstack->flags |= MCOUNT_FL_CALLER;
+
 		if (tr->flags & TRIGGER_FL_FINISH) {
 			record_trace_data(mtdp, rstack, NULL);
-			mcount_finish();
+			mcount_finish_trigger();
 			return;
 		}
 	}
@@ -683,6 +1038,8 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 				save_trigger_read(mtdp, rstack, tr->read, false);
 				rstack->flags |= MCOUNT_FL_READ;
 			}
+			if (mcount_watchpoints)
+				save_watchpoint(mtdp, rstack, mcount_watchpoints);
 
 			if (mtdp->nr_events) {
 				bool flush = false;
@@ -710,7 +1067,7 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 		if (tr->flags & FLAGS_TO_CHECK) {
 			if (tr->flags & TRIGGER_FL_RECOVER) {
 				mcount_rstack_restore(mtdp);
-				*rstack->parent_loc = (unsigned long) mcount_return;
+				*rstack->parent_loc = mcount_return_fn;
 				rstack->flags |= MCOUNT_FL_RECOVER;
 			}
 			if (tr->flags & (TRIGGER_FL_TRACE_ON | TRIGGER_FL_TRACE_OFF))
@@ -766,7 +1123,11 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp,
 			save_trigger_read(mtdp, rstack, tr.read, true);
 		}
 
-		if (rstack->end_time - rstack->start_time > time_filter ||
+		if (mcount_watchpoints)
+			save_watchpoint(mtdp, rstack, mcount_watchpoints);
+
+		if (((rstack->end_time - rstack->start_time > time_filter) &&
+		     (!mcount_has_caller || rstack->flags & MCOUNT_FL_CALLER)) ||
 		    rstack->flags & (MCOUNT_FL_WRITTEN | MCOUNT_FL_TRACE)) {
 			if (record_trace_data(mtdp, rstack, retval) < 0)
 				pr_err("error during record");
@@ -842,8 +1203,8 @@ mcount_arch_parent_location(struct symtabs *symtabs, unsigned long *parent_loc,
 }
 #endif
 
-int mcount_entry(unsigned long *parent_loc, unsigned long child,
-		 struct mcount_regs *regs)
+static int __mcount_entry(unsigned long *parent_loc, unsigned long child,
+			  struct mcount_regs *regs)
 {
 	enum filter_result filtered;
 	struct mcount_thread_data *mtdp;
@@ -858,7 +1219,7 @@ int mcount_entry(unsigned long *parent_loc, unsigned long child,
 			return -1;
 	}
 	else {
-		if (!mcount_guard_recursion(mtdp, false))
+		if (!mcount_guard_recursion(mtdp))
 			return -1;
 	}
 
@@ -900,7 +1261,7 @@ int mcount_entry(unsigned long *parent_loc, unsigned long child,
 	rstack->event_idx  = ARGBUF_SIZE;
 
 	/* hijack the return address of child */
-	*parent_loc = (unsigned long)mcount_return;
+	*parent_loc = mcount_return_fn;
 
 	/* restore return address of parent */
 	if (mcount_auto_recover)
@@ -911,41 +1272,75 @@ int mcount_entry(unsigned long *parent_loc, unsigned long child,
 	return 0;
 }
 
-unsigned long mcount_exit(long *retval)
+int mcount_entry(unsigned long *parent_loc, unsigned long child,
+		 struct mcount_regs *regs)
+{
+	int saved_errno = errno;
+	int ret = __mcount_entry(parent_loc, child, regs);
+
+	errno = saved_errno;
+	return ret;
+}
+
+static unsigned long __mcount_exit(long *retval)
 {
 	struct mcount_thread_data *mtdp;
 	struct mcount_ret_stack *rstack;
+	unsigned long *ret_loc;
 	unsigned long retaddr;
 
 	mtdp = get_thread_data();
 	assert(mtdp != NULL);
+	assert(!mtdp->dead);
 
 	/*
-	 * there's a race with mcount_finish(), but it still needs to get
-	 * the original return address so defer freeing rstack to the end.
+	 * it's only called when mcount_entry() was succeeded
+	 * no need to check recursion here.  But still needs to
+	 * prevent recursion during this call.
 	 */
-	mcount_guard_recursion(mtdp, true);
+	__mcount_guard_recursion(mtdp);
 
 	rstack = &mtdp->rstack[mtdp->idx - 1];
 
 	rstack->end_time = mcount_gettime();
 	mcount_exit_filter_record(mtdp, rstack, retval);
 
+	ret_loc = rstack->parent_loc;
 	retaddr = rstack->parent_ip;
 
 	/* re-hijack return address of parent */
 	if (mcount_auto_recover)
 		mcount_auto_reset(mtdp);
 
+	__mcount_unguard_recursion(mtdp);
+
+	if (unlikely(mcount_should_stop())) {
+		mtd_dtor(mtdp);
+		/*
+		 * mtd_dtor() will free rstack but current ret_addr
+		 * might be plthook_return() when it was a tailcall.
+		 * reload the return address after mtd_dtor() restored
+		 * all the parent locations.
+		 */
+		retaddr = *ret_loc;
+	}
+
 	compiler_barrier();
 
 	mtdp->idx--;
-	mcount_unguard_recursion(mtdp);
-
 	return retaddr;
 }
 
-static int cygprof_entry(unsigned long parent, unsigned long child)
+unsigned long mcount_exit(long *retval)
+{
+	int saved_errno = errno;
+	unsigned long ret = __mcount_exit(retval);
+
+	errno = saved_errno;
+	return ret;
+}
+
+static int __cygprof_entry(unsigned long parent, unsigned long child)
 {
 	enum filter_result filtered;
 	struct mcount_thread_data *mtdp;
@@ -962,7 +1357,7 @@ static int cygprof_entry(unsigned long parent, unsigned long child)
 			return -1;
 	}
 	else {
-		if (!mcount_guard_recursion(mtdp, false))
+		if (!mcount_guard_recursion(mtdp))
 			return -1;
 	}
 
@@ -1023,7 +1418,16 @@ static int cygprof_entry(unsigned long parent, unsigned long child)
 	return 0;
 }
 
-static void cygprof_exit(unsigned long parent, unsigned long child)
+static int cygprof_entry(unsigned long parent, unsigned long child)
+{
+	int saved_errno = errno;
+	int ret = __cygprof_entry(parent, child);
+
+	errno = saved_errno;
+	return ret;
+}
+
+static void __cygprof_exit(unsigned long parent, unsigned long child)
 {
 	struct mcount_thread_data *mtdp;
 	struct mcount_ret_stack *rstack;
@@ -1032,7 +1436,7 @@ static void cygprof_exit(unsigned long parent, unsigned long child)
 	if (unlikely(check_thread_data(mtdp)))
 		return;
 
-	if (!mcount_guard_recursion(mtdp, false))
+	if (!mcount_guard_recursion(mtdp))
 		return;
 
 	/*
@@ -1050,15 +1454,24 @@ static void cygprof_exit(unsigned long parent, unsigned long child)
 
 	mcount_exit_filter_record(mtdp, rstack, NULL);
 
+out:
+	mcount_unguard_recursion(mtdp);
+
 	compiler_barrier();
 
-out:
 	mtdp->idx--;
-	mcount_unguard_recursion(mtdp);
 }
 
-void xray_entry(unsigned long parent, unsigned long child,
-		struct mcount_regs *regs)
+static void cygprof_exit(unsigned long parent, unsigned long child)
+{
+	int saved_errno = errno;
+
+	__cygprof_exit(parent, child);
+	errno = saved_errno;
+}
+
+static void _xray_entry(unsigned long parent, unsigned long child,
+			struct mcount_regs *regs)
 {
 	enum filter_result filtered;
 	struct mcount_thread_data *mtdp;
@@ -1075,7 +1488,7 @@ void xray_entry(unsigned long parent, unsigned long child,
 			return;
 	}
 	else {
-		if (!mcount_guard_recursion(mtdp, false))
+		if (!mcount_guard_recursion(mtdp))
 			return;
 	}
 
@@ -1123,7 +1536,16 @@ void xray_entry(unsigned long parent, unsigned long child,
 	mcount_unguard_recursion(mtdp);
 }
 
-void xray_exit(long *retval)
+void xray_entry(unsigned long parent, unsigned long child,
+		struct mcount_regs *regs)
+{
+	int saved_errno = errno;
+
+	_xray_entry(parent, child, regs);
+	errno = saved_errno;
+}
+
+static void _xray_exit(long *retval)
 {
 	struct mcount_thread_data *mtdp;
 	struct mcount_ret_stack *rstack;
@@ -1132,7 +1554,7 @@ void xray_exit(long *retval)
 	if (unlikely(check_thread_data(mtdp)))
 		return;
 
-	if (!mcount_guard_recursion(mtdp, false))
+	if (!mcount_guard_recursion(mtdp))
 		return;
 
 	/*
@@ -1150,35 +1572,20 @@ void xray_exit(long *retval)
 
 	mcount_exit_filter_record(mtdp, rstack, retval);
 
+out:
+	mcount_unguard_recursion(mtdp);
+
 	compiler_barrier();
 
-out:
 	mtdp->idx--;
-	mcount_unguard_recursion(mtdp);
 }
 
-/* save an asynchronous event */
-int mcount_save_event(struct mcount_event_info *mei)
+void xray_exit(long *retval)
 {
-	struct mcount_thread_data *mtdp;
+	int saved_errno = errno;
 
-	if (unlikely(mcount_should_stop()))
-		return -1;
-
-	mtdp = get_thread_data();
-	if (unlikely(check_thread_data(mtdp)))
-		return -1;
-
-	if (mtdp->nr_events < MAX_EVENT) {
-		int i = mtdp->nr_events++;
-
-		mtdp->event[i].id   = mei->id;
-		mtdp->event[i].time = mcount_gettime();
-		mtdp->event[i].dsize = 0;
-		mtdp->event[i].idx   = ASYNC_IDX;
-	}
-
-	return 0;
+	_xray_exit(retval);
+	errno = saved_errno;
 }
 
 static void atfork_prepare_handler(void)
@@ -1193,6 +1600,10 @@ static void atfork_prepare_handler(void)
 		script_atfork_prepare();
 
 	uftrace_send_message(UFTRACE_MSG_FORK_START, &tmsg, sizeof(tmsg));
+
+	/* flush remaining contents in the stream */
+	fflush(outfp);
+	fflush(logfp);
 }
 
 static void atfork_child_handler(void)
@@ -1203,18 +1614,16 @@ static void atfork_child_handler(void)
 		.pid = getppid(),
 		.tid = getpid(),
 	};
+	int i;
 
 	mtdp = get_thread_data();
 	if (unlikely(check_thread_data(mtdp))) {
-		/* we need it even if in a recursion */
-		mtdp->recursion_marker = false;
-
 		mtdp = mcount_prepare();
 		if (mtdp == NULL)
 			return;
 	}
 	else {
-		if (!mcount_guard_recursion(mtdp, false))
+		if (!mcount_guard_recursion(mtdp))
 			return;
 	}
 
@@ -1230,6 +1639,10 @@ static void atfork_child_handler(void)
 
 	update_kernel_tid(tmsg.tid);
 
+	/* do not record parent's functions */
+	for (i = 0; i < mtdp->idx; i++)
+		mtdp->rstack[i].flags |= MCOUNT_FL_WRITTEN;
+
 	mcount_unguard_recursion(mtdp);
 }
 
@@ -1238,21 +1651,21 @@ static void mcount_script_init(enum uftrace_pattern_type patt_type)
 	struct script_info info = {
 		.name           = script_str,
 		.version        = UFTRACE_VERSION,
-		.recording      = true,
+		.record         = true,
 	};
-	char *args_str;
+	char *cmds_str;
 
-	args_str = getenv("UFTRACE_ARGS");
-	if (args_str)
-		strv_split(&info.args, args_str, "\n");
+	cmds_str = getenv("UFTRACE_ARGS");
+	if (cmds_str)
+		strv_split(&info.cmds, cmds_str, "\n");
 
 	if (script_init(&info, patt_type) < 0)
 		script_str = NULL;
 
-	strv_free(&info.args);
+	strv_free(&info.cmds);
 }
 
-static void mcount_startup(void)
+static __used void mcount_startup(void)
 {
 	char *pipefd_str;
 	char *logfd_str;
@@ -1321,12 +1734,16 @@ static void mcount_startup(void)
 		demangler = strtol(demangle_str, NULL, 0);
 
 	if (color_str)
-		setup_color(strtol(color_str, NULL, 0));
+		setup_color(strtol(color_str, NULL, 0), NULL);
 	else
-		setup_color(COLOR_AUTO);
+		setup_color(COLOR_AUTO, NULL);
 
 	pr_dbg("initializing mcount library\n");
 
+	dirname = getenv("UFTRACE_DIR");
+	if (dirname == NULL)
+		dirname = UFTRACE_DIR_NAME;
+
 	if (pipefd_str) {
 		pfd = strtol(pipefd_str, NULL, 0);
 
@@ -1336,6 +1753,13 @@ static void mcount_startup(void)
 			pfd = -1;
 		}
 	}
+	else {
+		char *channel = NULL;
+
+		xasprintf(&channel, "%s/%s", dirname, ".channel");
+		pfd = open(channel, O_WRONLY);
+		free(channel);
+	}
 
 	if (getenv("UFTRACE_LIST_EVENT")) {
 		mcount_list_events();
@@ -1345,21 +1769,22 @@ static void mcount_startup(void)
 	if (bufsize_str)
 		shmem_bufsize = strtol(bufsize_str, NULL, 0);
 
-	dirname = getenv("UFTRACE_DIR");
-	if (dirname == NULL)
-		dirname = UFTRACE_DIR_NAME;
-
 	mcount_exename = read_exename();
 	symtabs.dirname = dirname;
 	symtabs.filename = mcount_exename;
 
 	record_proc_maps(dirname, mcount_session_name(), &symtabs);
-	load_symtabs(&symtabs, NULL, mcount_exename);
 
 	if (pattern_str)
 		patt_type = parse_filter_pattern(pattern_str);
 
+	if (patch_str)
+		mcount_return_fn = (unsigned long)dynamic_return;
+	else
+		mcount_return_fn = (unsigned long)mcount_return;
+
 	mcount_filter_init(patt_type, dirname, !!patch_str);
+	mcount_watch_init();
 
 	if (maxstack_str)
 		mcount_rstack_max = strtol(maxstack_str, NULL, 0);
@@ -1368,7 +1793,7 @@ static void mcount_startup(void)
 		mcount_threshold = strtoull(threshold_str, NULL, 0);
 
 	if (patch_str)
-		mcount_dynamic_update(&symtabs, patch_str, patt_type);
+		mcount_dynamic_update(&symtabs, patch_str, patt_type, &disasm);
 
 	if (event_str)
 		mcount_setup_events(dirname, event_str, patt_type);
@@ -1399,19 +1824,23 @@ static void mcount_cleanup(void)
 	mcount_finish();
 	destroy_dynsym_indexes();
 
+#if 0
+	/*
+	 * This mtd_key deletion sometimes makes other thread get crashed
+	 * because they may try to get mtdp based on this mtd_key after being
+	 * deleted.  Since this key deletion is not mandatory, it'd be better
+	 * not to delete it until we find a better solution.
+	 */
 	pthread_key_delete(mtd_key);
 	mtd_key = -1;
-
-#ifndef DISABLE_MCOUNT_FILTER
-	uftrace_cleanup_filter(&mcount_triggers);
 #endif
+
+	mcount_filter_finish();
+
 	if (SCRIPT_ENABLED && script_str)
 		script_finish();
 
-	finish_debug_info(&symtabs);
-	unload_symtabs(&symtabs);
-	finish_pmu_event();
-	finish_auto_args();
+	unload_module_symtabs();
 
 	pr_dbg("exit from libmcount\n");
 }
@@ -1493,10 +1922,7 @@ TEST_CASE(mcount_thread_data)
 {
 	struct mcount_thread_data *mtdp;
 
-	if (0)
-		mcount_startup();
-	else
-		setup_mcount_test();
+	setup_mcount_test();
 
 	mtdp = get_thread_data();
 	TEST_EQ(check_thread_data(mtdp), true);
@@ -1513,4 +1939,34 @@ TEST_CASE(mcount_thread_data)
 	return TEST_OK;
 }
 
+TEST_CASE(mcount_signal_setup)
+{
+	struct signal_trigger_item *item;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_NONE,
+	};
+
+	/* it signal triggers are maintained in a stack (LIFO) */
+	mcount_signal_init("SIGUSR1@traceon;USR2@traceoff;RTMIN+3@finish",
+			   &setting);
+
+	item = list_first_entry(&siglist, typeof(*item), list);
+	TEST_EQ(item->sig, SIGRTMIN + 3);
+	TEST_EQ(item->tr.flags, TRIGGER_FL_FINISH);
+
+	item = list_next_entry(item, list);
+	TEST_EQ(item->sig, SIGUSR2);
+	TEST_EQ(item->tr.flags, TRIGGER_FL_TRACE_OFF);
+
+	item = list_next_entry(item, list);
+	TEST_EQ(item->sig, SIGUSR1);
+	TEST_EQ(item->tr.flags, TRIGGER_FL_TRACE_ON);
+
+	mcount_signal_finish();
+
+	TEST_EQ(list_empty(&siglist), true);
+
+	return TEST_OK;
+}
+
 #endif /* UNIT_TEST */
