@@ -11,12 +11,20 @@ int main(int argc, char *argv[])
 	if (argc > 1)
 		n = atoi(argv[1]);
 
-	handle = dlopen("libabc_test_lib.so", RTLD_LAZY);
+	handle = dlopen("./libabc_test_lib.so", RTLD_LAZY);
+
+	if (!handle)
+		return -1;
+
 	sym = dlsym(handle, "lib_a");
 	sym(n);
 	dlclose(handle);
 
-	handle = dlopen("libfoo.so", RTLD_LAZY);
+	handle = dlopen("./libfoo.so", RTLD_LAZY);
+
+	if (!handle)
+		return -1;
+
 	sym = dlsym(handle, "foo");
 	sym(n);
 	dlclose(handle);
