@@ -7,7 +7,8 @@
 #include <unistd.h>
 #include <sys/wait.h>
 
-#define TEST_PROG  "t-abc"
+#define TEST_PROG1  "t-abc"
+#define TEST_PROG2  "t-openclose"
 
 int main(int argc, char *argv[])
 {
@@ -15,7 +16,10 @@ int main(int argc, char *argv[])
 
 	pid = fork();
 	if (pid == 0) {
-		execl(TEST_PROG, TEST_PROG, NULL);
+		if (argc == 1)
+			execl(TEST_PROG1, TEST_PROG1, NULL);
+		else
+			execl(TEST_PROG2, TEST_PROG2, NULL);
 		exit(2);
 	}
 	waitpid(pid, NULL, 0);
