@@ -3,6 +3,7 @@
 #include <stdlib.h>
 
 void * (*real_malloc)(size_t sz);
+void * (*real_realloc)(void *ptr, size_t sz);
 void (*real_free)(void *ptr);
 
 #define ALIGN(n, a)  (((n) + (a) - 1) & ~((a) - 1))
@@ -29,6 +30,27 @@ void *malloc(size_t sz)
 	return ptr;
 }
 
+void *realloc(void *ptr, size_t size)
+{
+	char *p;
+
+	if (real_realloc && (ptr < buf || ptr >= &buf[MALLOC_BUFSIZE]))
+		return real_realloc(ptr, size);
+
+	p = malloc(size);
+
+	/* using memcpy() caused segfault due to alignment */
+	if (ptr != NULL) {
+		char *q = ptr;
+		size_t i;
+
+		for (i = 0; i < size; i++)
+			p[i] = q[i];
+	}
+
+	return p;
+}
+
 void free(void *ptr)
 {
 	char *p = ptr;
@@ -42,8 +64,9 @@ void free(void *ptr)
 
 static void hook(void)
 {
-	real_malloc = dlsym(RTLD_NEXT, "malloc");
-	real_free   = dlsym(RTLD_NEXT, "free");
+	real_malloc  = dlsym(RTLD_NEXT, "malloc");
+	real_realloc = dlsym(RTLD_NEXT, "realloc");
+	real_free    = dlsym(RTLD_NEXT, "free");
 }
 
 static __attribute__((section(".preinit_array")))
