@@ -13,7 +13,7 @@ void *malloc(size_t size)
 	static unsigned alloc_size;
 	void *ptr;
 
-	size = ALIGN(size, 8);
+	size = ALIGN(size, 16);
 	if (alloc_size + size > sizeof(buf))
 		return NULL;
 
@@ -24,6 +24,15 @@ void *malloc(size_t size)
 	return ptr;
 }
 
+void *realloc(void *ptr, size_t size)
+{
+	void *p = malloc(size);
+
+	if (ptr)
+		memcpy(p, ptr, size);
+	return p;
+}
+
 void free(void *ptr)
 {
 	free_count++;
