@@ -26,4 +26,4 @@ class TestCase(TestBase):
 
     # test whether filter option preserves the ordering
     def runcmd(self):
-        return '%s -F "ns1::.*" -N "bar2$" -F "bar2$" %s' % (TestBase.ftrace, 't-namespace')
+        return '%s -F "ns1::.*" -N "bar2$" -F "bar2$" %s' % (TestBase.uftrace_cmd, 't-namespace')
