@@ -23,4 +23,4 @@ class TestCase(TestBase):
         return TestBase.build(self, name, cflags, ldflags)
 
     def runcmd(self):
-        return '%s -A "^str_@arg1/s,arg2/s" %s' % (TestBase.ftrace, 't-' + self.name)
+        return '%s -A "^str_@arg1/s,arg2/s" %s' % (TestBase.uftrace_cmd, 't-' + self.name)
