@@ -21,12 +21,12 @@ class TestCase(TestBase):
                                       ['libabc_test_lib.so'])
 
     def pre(self):
-        record_cmd = '%s record --force -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
+        record_cmd = '%s record --force -d %s %s' % (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s replay -d %s -F lib_b@libabc_test' % (TestBase.ftrace, TDIR)
+        return '%s replay -d %s -F lib_b@libabc_test' % (TestBase.uftrace_cmd, TDIR)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
