@@ -22,7 +22,7 @@ class TestCase(TestBase):
 """, sort='report')
 
     def pre(self):
-        record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
+        record_cmd = '%s record -d %s %s' % (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
@@ -38,7 +38,7 @@ class TestCase(TestBase):
                 pass
         if t == 0:
             return 'FAILED TO FIND TID'
-        return '%s report -d %s --tid %d' % (TestBase.ftrace, TDIR, t)
+        return '%s report -d %s --tid %d' % (TestBase.uftrace_cmd, TDIR, t)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
