@@ -9,6 +9,8 @@ class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'abc', """
 {"traceEvents":[
+{"ts":0,"ph":"M","pid":5231,"name":"process_name","args":{"name":"[5231] t-abc"}},
+{"ts":0,"ph":"M","pid":5231,"name":"thread_name","args":{"name":"[5231] t-abc"}},
 {"ts":58348873444,"ph":"B","pid":5231,"name":"main"},
 {"ts":58348873444,"ph":"B","pid":5231,"name":"a"},
 {"ts":58348873445,"ph":"B","pid":5231,"name":"b"},
@@ -24,12 +26,12 @@ class TestCase(TestBase):
 """, sort='chrome')
 
     def pre(self):
-        record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
+        record_cmd = '%s record -d %s %s' % (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s dump -d %s -F main -D 4 --chrome' % (TestBase.ftrace, TDIR)
+        return '%s dump -d %s -F main -D 4 --chrome' % (TestBase.uftrace_cmd, TDIR)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
