@@ -2,16 +2,16 @@
 
 from runtest import TestBase
 import subprocess as sp
-import os, re
+import os
 
 TDIR='xxx'
 
 class TestCase(TestBase):
     def __init__(self):
-        TestBase.__init__(self, 'getids', """
+        TestBase.__init__(self, 'getids', serial=True, result="""
 {"traceEvents":[
-{"ts":14510734170,"ph":"M","pid":32687,"name":"process_name","args":"{'name': 't-getids'}"},
-{"ts":14510734170,"ph":"M","pid":32687,"name":"thread_name","args":"{'name': 't-getids'}"},
+{"ts":14510734170,"ph":"M","pid":32687,"name":"process_name","args":{'name': 't-getids'}},
+{"ts":14510734170,"ph":"M","pid":32687,"name":"thread_name","args":{'name': 't-getids'}},
 {"ts":14510734172,"ph":"B","pid":32687,"name":"main"},
 {"ts":14510734172,"ph":"B","pid":32687,"name":"getpid"},
 {"ts":14510734173,"ph":"E","pid":32687,"name":"getpid"},
@@ -56,8 +56,8 @@ class TestCase(TestBase):
         if os.path.exists('/.dockerenv'):
             return TestBase.TEST_SKIP
 
-        record_cmd = '%s record -k -N %s@kernel -d %s %s' % \
-                     (TestBase.uftrace_cmd, 'smp_irq_work_interrupt', TDIR, 't-' + self.name)
+        record_cmd = '%s record -k -d %s %s' % \
+                     (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
@@ -85,6 +85,6 @@ class TestCase(TestBase):
         major, minor, release = uname[2].split('.')
         if uname[0] == 'Linux' and uname[4] == 'x86_64' and \
            int(major) >= 4 and int(minor) >= 17:
-            result = re.sub('sys_[a-zA-Z0-9_]+', 'do_syscall_64', result)
+            result = result.replace('sys_get', '__x64_sys_get')
 
         return result
