@@ -22,12 +22,12 @@ class TestCase(TestBase):
 """)
 
     def pre(self):
-        record_cmd = "%s record -d %s %s" % (TestBase.ftrace, TDIR, 't-' + self.name)
+        record_cmd = "%s record -d %s %s" % (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s replay -F main -d %s' % (TestBase.ftrace, TDIR)
+        return '%s replay -F main -d %s' % (TestBase.uftrace_cmd, TDIR)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
