@@ -13,4 +13,4 @@ class TestCase(TestBase):
 """)
 
     def runcmd(self):
-        return '%s -F main %s' % (TestBase.ftrace, 't-' + self.name)
+        return '%s -F main %s' % (TestBase.uftrace_cmd, 't-' + self.name)
