@@ -9,9 +9,7 @@ class TestCase(TestBase):
 # DURATION    TID     FUNCTION
   62.202 us [28141] | __cxa_atexit();
             [28141] | main() {
-            [28141] |   a() {
-   0.753 us [28141] |     getpid();
-   2.405 us [28141] |   } /* a */
+   2.405 us [28141] |   a();
    3.005 us [28141] | } /* main */
 """)
 
@@ -21,7 +19,12 @@ class TestCase(TestBase):
         return TestBase.TEST_SUCCESS
 
     def build(self, name, cflags='', ldflags=''):
-        return TestBase.build(self, name, '-pg -mfentry -mnop-mcount', ldflags)
+        cflags += ' -mfentry -mnop-mcount'
+        cflags += ' -fno-pie -fno-plt'  # workaround of build failure
+        return TestBase.build(self, name, cflags, ldflags)
 
     def runcmd(self):
-        return '%s -P %s %s' % (TestBase.ftrace, 'a.?', 't-' + self.name)
+        uftrace  = TestBase.uftrace_cmd
+        argument = '-P %s --no-libcall' % 'a.?'
+        program  = 't-' + self.name
+        return '%s %s %s' % (uftrace, argument, program)
