@@ -9,17 +9,18 @@ class TestCase(TestBase):
             [32417] | main() {
             [32417] |   a() {
             [32417] |     b() {
-            [32417] |       /* read:proc/statm (size=6810KB, rss=785KB, shared=719KB) */
+            [32417] |       /* read:proc/statm (size=6812KB, rss=780KB, shared=716KB) */
             [32417] |       c() {
    0.479 us [32417] |         getpid();
    3.014 us [32417] |       } /* c */
+            [32417] |       /* diff:proc/statm (size=+0KB, rss=+0KB, shared=+0KB) */
   16.914 us [32417] |     } /* b */
   17.083 us [32417] |   } /* a */
   17.873 us [32417] | } /* main */
 """)
 
     def runcmd(self):
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         args    = '-F main -T b@read=proc/statm'
         prog    = 't-' + self.name
         return '%s %s %s' % (uftrace, args, prog)
@@ -34,6 +35,8 @@ class TestCase(TestBase):
             # remove actual numbers in proc.statm
             if func.find('read:proc/statm') > 0:
                 func = '       /* read:proc/statm */'
+            if func.find('diff:proc/statm') > 0:
+                func = '       /* diff:proc/statm */'
             result.append(func)
 
         return '\n'.join(result)
