@@ -13,6 +13,7 @@ class TestCase(TestBase):
             [ 6565] |     ns::ns1::foo::bar1() {
             [ 6565] |       ns::ns1::foo::bar2() {
             [ 6565] |         ns::ns1::foo::bar3() {
+            [ 6565] |           /* linux:task-exit */
 
 uftrace stopped tracing with remaining functions
 ================================================
@@ -25,7 +26,12 @@ task: 6565
 """, lang='C++')
 
     def runcmd(self):
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         options = '-F main -T ns::ns1::foo::bar3@finish'
         program = 't-' + self.name
         return '%s %s %s' % (uftrace, options, program)
+
+    def fixup(self, cflags, result):
+        return result.replace("""ns::ns1::foo::bar3() {
+            [ 6565] |           /* linux:task-exit */""",
+                                "ns::ns1::foo::bar3() {")
