@@ -26,11 +26,16 @@ a exit
 """)
 
     def pre(self):
+        script_cmd = '%s script' % (TestBase.uftrace_cmd)
+        p = sp.Popen(script_cmd.split(), stdout=sp.PIPE, stderr=sp.PIPE)
+        if p.communicate()[1].decode(errors='ignore').startswith('WARN:'):
+            return TestBase.TEST_SKIP
+
         f = open(FILE, 'w')
         f.write(script)
         f.close()
 
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         program = 't-' + self.name
         record_cmd = '%s record -d %s %s' % (uftrace, TDIR, program)
 
@@ -39,7 +44,7 @@ a exit
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         options = '-S ' + FILE
         return '%s script -d %s %s' % (uftrace, TDIR, options)
 
