@@ -20,7 +20,7 @@ class TestCase(TestBase):
 """)
 
     def runcmd(self):
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         options = '-T main@filter,depth=3 -T ^ns::ns2@notrace'
         program = 't-' + self.name
         return '%s %s %s' % (uftrace, options, program)
