@@ -32,7 +32,7 @@ class TestCase(TestBase):
         return TestBase.build(self, name, cflags, ldflags)
 
     def runcmd(self):
-        uftrace = TestBase.ftrace
+        uftrace = TestBase.uftrace_cmd
         options = '-T main@filter,depth=3,arg1'
         program = 't-' + self.name
         return '%s %s %s' % (uftrace, options, program)
