@@ -22,7 +22,8 @@ static struct rb_root auto_enum = RB_ROOT;
 extern void add_trigger(struct uftrace_filter *filter, struct uftrace_trigger *tr,
 			bool exact_match);
 extern int setup_trigger_action(char *str, struct uftrace_trigger *tr,
-				char **module, unsigned long orig_flags);
+				char **module, unsigned long orig_flags,
+				struct uftrace_filter_setting *setting);
 
 static void add_auto_args(struct rb_root *root, struct uftrace_filter *entry,
 			  struct uftrace_trigger *tr)
@@ -63,7 +64,8 @@ static void add_auto_args(struct rb_root *root, struct uftrace_filter *entry,
 }
 
 static void build_auto_args(const char *args_str, struct rb_root *root,
-			    unsigned long flag)
+			    unsigned long flag,
+			    struct uftrace_filter_setting *setting)
 {
 	struct strv specs = STRV_INIT;
 	char *name;
@@ -95,7 +97,7 @@ static void build_auto_args(const char *args_str, struct rb_root *root,
 		 */
 		entry.end = (unsigned long)xstrdup(p + 1);
 
-		if (setup_trigger_action(name, &tr, NULL, flag) < 0)
+		if (setup_trigger_action(name, &tr, NULL, flag, setting) < 0)
 			goto next;
 
 		/*
@@ -144,7 +146,8 @@ static struct uftrace_filter *dwarf_argspec_list;
 
 static struct uftrace_filter * find_dwarf_argspec(struct uftrace_filter *filter,
 						  struct debug_info *dinfo,
-						  bool is_retval)
+						  bool is_retval,
+						  struct uftrace_filter_setting *setting)
 {
 	LIST_HEAD(dwarf_argspec);
 	struct uftrace_filter *dwarf_filter;
@@ -153,17 +156,21 @@ static struct uftrace_filter * find_dwarf_argspec(struct uftrace_filter *filter,
 	};
 	char *arg_str;
 	unsigned long flag = is_retval ? TRIGGER_FL_RETVAL : TRIGGER_FL_ARGUMENT;
+	unsigned long addr = filter->start;
 
 	if (is_retval)
-		arg_str = get_dwarf_retspec(dinfo, filter->name, filter->start);
+		arg_str = get_dwarf_retspec(dinfo, filter->name, addr);
 	else
-		arg_str = get_dwarf_argspec(dinfo, filter->name, filter->start);
+		arg_str = get_dwarf_argspec(dinfo, filter->name, addr);
 	if (arg_str == NULL)
 		return NULL;
 
-	setup_trigger_action(arg_str, &dwarf_tr, NULL, flag);
-	if (list_empty(dwarf_tr.pargs))
+	arg_str = xstrdup(arg_str);
+	setup_trigger_action(arg_str, &dwarf_tr, NULL, flag, setting);
+	if (list_empty(dwarf_tr.pargs)) {
+		free(arg_str);
 		return NULL;
+	}
 
 	dwarf_filter = xzalloc(sizeof(*dwarf_filter));
 	INIT_LIST_HEAD(&dwarf_filter->args);
@@ -176,17 +183,19 @@ static struct uftrace_filter * find_dwarf_argspec(struct uftrace_filter *filter,
 	dwarf_filter->name = (void *)dwarf_argspec_list;
 	dwarf_argspec_list = dwarf_filter;
 
+	free(arg_str);
 	return dwarf_filter;
 }
 
 struct uftrace_filter * find_auto_argspec(struct uftrace_filter *filter,
 					  struct uftrace_trigger *tr,
-					  struct debug_info *dinfo)
+					  struct debug_info *dinfo,
+					  struct uftrace_filter_setting *setting)
 {
 	struct uftrace_filter *auto_arg = NULL;
 
 	if (debug_info_has_argspec(dinfo))
-		auto_arg = find_dwarf_argspec(filter, dinfo, false);
+		auto_arg = find_dwarf_argspec(filter, dinfo, false, setting);
 
 	if (auto_arg == NULL)
 		auto_arg = find_auto_args(&auto_argspec, filter->name);
@@ -196,12 +205,13 @@ struct uftrace_filter * find_auto_argspec(struct uftrace_filter *filter,
 
 struct uftrace_filter * find_auto_retspec(struct uftrace_filter *filter,
 					  struct uftrace_trigger *tr,
-					  struct debug_info *dinfo)
+					  struct debug_info *dinfo,
+					  struct uftrace_filter_setting *setting)
 {
 	struct uftrace_filter *auto_ret = NULL;
 
 	if (debug_info_has_argspec(dinfo))
-		auto_ret = find_dwarf_argspec(filter, dinfo, true);
+		auto_ret = find_dwarf_argspec(filter, dinfo, true, setting);
 
 	if (auto_ret == NULL)
 		auto_ret = find_auto_args(&auto_retspec, filter->name);
@@ -219,18 +229,22 @@ char *get_auto_retspec_str(void)
 	return auto_retvals_list;
 }
 
-void setup_auto_args(void)
+void setup_auto_args(struct uftrace_filter_setting *setting)
 {
+
 	parse_enum_string(auto_enum_list, &auto_enum);
-	build_auto_args(auto_args_list, &auto_argspec, TRIGGER_FL_ARGUMENT);
-	build_auto_args(auto_retvals_list, &auto_retspec, TRIGGER_FL_RETVAL);
+	build_auto_args(auto_args_list, &auto_argspec, TRIGGER_FL_ARGUMENT,
+			setting);
+	build_auto_args(auto_retvals_list, &auto_retspec, TRIGGER_FL_RETVAL,
+			setting);
 }
 
-void setup_auto_args_str(char *args, char *rets, char *enums)
+void setup_auto_args_str(char *args, char *rets, char *enums,
+			 struct uftrace_filter_setting *setting)
 {
 	parse_enum_string(enums, &auto_enum);
-	build_auto_args(args, &auto_argspec, TRIGGER_FL_ARGUMENT);
-	build_auto_args(rets, &auto_retspec, TRIGGER_FL_RETVAL);
+	build_auto_args(args, &auto_argspec, TRIGGER_FL_ARGUMENT, setting);
+	build_auto_args(rets, &auto_retspec, TRIGGER_FL_RETVAL, setting);
 }
 
 static void release_auto_args(struct rb_root *root)
@@ -439,6 +453,7 @@ static void free_enum_def(struct enum_def *e_def)
 		free(e_val->str);
 		free(e_val);
 	}
+	free(e_def->name);
 	free(e_def);
 }
 
@@ -749,12 +764,16 @@ TEST_CASE(argspec_auto_args)
 	struct uftrace_filter *entry;
 	struct uftrace_filter key;
 	struct uftrace_arg_spec *spec;
+	struct uftrace_filter_setting setting = {
+		.lp64 = host_is_lp64(),
+	};
 	int idx = 1;
 
-	build_auto_args(test_auto_args, &auto_argspec, TRIGGER_FL_ARGUMENT);
+	build_auto_args(test_auto_args, &auto_argspec, TRIGGER_FL_ARGUMENT,
+			&setting);
 
 	key.name = "foo";
-	entry = find_auto_argspec(&key, NULL, NULL);
+	entry = find_auto_argspec(&key, NULL, NULL, &setting);
 	TEST_NE(entry, NULL);
 	TEST_EQ(entry->trigger.flags, TRIGGER_FL_ARGUMENT);
 
@@ -765,7 +784,7 @@ TEST_CASE(argspec_auto_args)
 	}
 
 	key.name = "bar";
-	entry = find_auto_argspec(&key, NULL, NULL);
+	entry = find_auto_argspec(&key, NULL, NULL, &setting);
 	TEST_NE(entry, NULL);
 	TEST_EQ(entry->trigger.flags, TRIGGER_FL_ARGUMENT);
 
@@ -774,13 +793,13 @@ TEST_CASE(argspec_auto_args)
 	TEST_EQ(spec->idx, 1);
 
 	key.name = "xxx";
-	entry = find_auto_argspec(&key, NULL, NULL);
+	entry = find_auto_argspec(&key, NULL, NULL, &setting);
 	TEST_EQ(entry, NULL);
 
 	release_auto_args(&auto_argspec);
 
 	key.name = "foo";
-	entry = find_auto_argspec(&key, NULL, NULL);
+	entry = find_auto_argspec(&key, NULL, NULL, &setting);
 	TEST_EQ(entry, NULL);
 
 	return TEST_OK;
