@@ -6,6 +6,8 @@
 #include <errno.h>
 #include <assert.h>
 #include <byteswap.h>
+#include <glob.h>
+#include <sys/stat.h>
 
 #include "uftrace.h"
 #include "utils/utils.h"
@@ -21,8 +23,9 @@
  * read_task_file - read 'task' file from data directory
  * @sess: session link to manage sessions and tasks
  * @dirname: name of the data directory
- * @needs_session: read session info too
+ * @needs_symtab: read session symbol tables too
  * @sym_rel_addr: whether symbol address is relative
+ * @needs_srcline: whether debug info loading is needed
  *
  * This function read the task file in the @dirname and build task
  * (and session when @needs_session is %true) information.  Note that
@@ -32,7 +35,7 @@
  * It returns 0 for success, -1 for error.
  */
 int read_task_file(struct uftrace_session_link *sess, char *dirname,
-		   bool needs_session, bool sym_rel_addr)
+		   bool needs_symtab, bool sym_rel_addr, bool needs_srcline)
 {
 	int fd;
 	char pad[8];
@@ -62,22 +65,23 @@ int read_task_file(struct uftrace_session_link *sess, char *dirname,
 			    read_all(fd, pad, 8 - (smsg.namelen % 8)) < 0)
 				goto out;
 
-			if (needs_session)
-				create_session(sess, &smsg, dirname, buf, sym_rel_addr);
+			create_session(sess, &smsg, dirname, buf,
+				       sym_rel_addr, needs_symtab,
+				       needs_srcline);
 			break;
 
 		case UFTRACE_MSG_TASK_START:
 			if (read_all(fd, &tmsg, sizeof(tmsg)) < 0)
 				goto out;
 
-			create_task(sess, &tmsg, false, needs_session);
+			create_task(sess, &tmsg, false);
 			break;
 
 		case UFTRACE_MSG_FORK_END:
 			if (read_all(fd, &tmsg, sizeof(tmsg)) < 0)
 				goto out;
 
-			create_task(sess, &tmsg, true, needs_session);
+			create_task(sess, &tmsg, true);
 			break;
 
 		default:
@@ -96,8 +100,9 @@ out:
  * read_task_txt_file - read 'task.txt' file from data directory
  * @sess: session link to manage sessions and tasks
  * @dirname: name of the data directory
- * @needs_session: read session info too
+ * @needs_symtab: read session symbol tables too
  * @sym_rel_addr: whethere symbol address is relative
+ * @needs_srcline: whether debug info loading is needed
  *
  * This function read the task.txt file in the @dirname and build task
  * (and session when @needs_session is %true) information.
@@ -105,7 +110,7 @@ out:
  * It returns 0 for success, -1 for error.
  */
 int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
-		       bool needs_session, bool sym_rel_addr)
+		       bool needs_symtab, bool sym_rel_addr, bool needs_srcline)
 {
 	FILE *fp;
 	char *fname = NULL;
@@ -116,6 +121,8 @@ int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
 	struct uftrace_msg_sess smsg;
 	struct uftrace_msg_dlopen dlop;
 	char *exename, *pos;
+	int ret = -1;
+	int num;
 
 	xasprintf(&fname, "%s/%s", dirname, "task.txt");
 
@@ -128,30 +135,34 @@ int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
 	pr_dbg("reading %s file\n", fname);
 	while (getline(&line, &sz, fp) >= 0) {
 		if (!strncmp(line, "TASK", 4)) {
-			sscanf(line + 5, "timestamp=%lu.%lu tid=%d pid=%d",
-			       &sec, &nsec, &tmsg.tid, &tmsg.pid);
+			num = sscanf(line + 5, "timestamp=%lu.%lu tid=%d pid=%d",
+				     &sec, &nsec, &tmsg.tid, &tmsg.pid);
+			if (num != 4)
+				goto out;
 
 			tmsg.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
-			create_task(sess, &tmsg, false, needs_session);
+			create_task(sess, &tmsg, false);
 		}
 		else if (!strncmp(line, "FORK", 4)) {
-			sscanf(line + 5, "timestamp=%lu.%lu pid=%d ppid=%d",
-			       &sec, &nsec, &tmsg.tid, &tmsg.pid);
+			num = sscanf(line + 5, "timestamp=%lu.%lu pid=%d ppid=%d",
+				     &sec, &nsec, &tmsg.tid, &tmsg.pid);
+			if (num != 4)
+				goto out;
 
 			tmsg.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
-			create_task(sess, &tmsg, true, needs_session);
+			create_task(sess, &tmsg, true);
 		}
 		else if (!strncmp(line, "SESS", 4)) {
-			if (!needs_session)
-				continue;
-
-			sscanf(line + 5, "timestamp=%lu.%lu %*[^i]id=%d sid=%s",
-			       &sec, &nsec, &smsg.task.pid, (char *)&smsg.sid);
+			num = sscanf(line + 5, "timestamp=%lu.%lu %*[^i]id=%d sid=%s",
+				     &sec, &nsec, &smsg.task.pid, (char *)&smsg.sid);
+			if (num != 4)
+				goto out;
 
 			// Get the execname
 			pos = strstr(line, "exename=");
 			if (pos == NULL)
-				pr_err_ns("invalid task.txt format");
+				goto out;
+
 			exename = pos + 8 + 1;  // skip double-quote
 			pos = strrchr(exename, '\"');
 			if (pos)
@@ -161,21 +172,26 @@ int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
 			smsg.task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
 			smsg.namelen = strlen(exename);
 
-			create_session(sess, &smsg, dirname, exename, sym_rel_addr);
+			create_session(sess, &smsg, dirname, exename,
+				       sym_rel_addr, needs_symtab,
+				       needs_srcline);
 		}
 		else if (!strncmp(line, "DLOP", 4)) {
 			struct uftrace_session *s;
 
-			if (!needs_session)
+			if (!needs_symtab)
 				continue;
 
-			sscanf(line + 5, "timestamp=%lu.%lu tid=%d sid=%s base=%"PRIx64,
-			       &sec, &nsec, &dlop.task.tid, (char *)&dlop.sid,
-			       &dlop.base_addr);
+			num = sscanf(line + 5, "timestamp=%lu.%lu tid=%d sid=%s base=%"PRIx64,
+				     &sec, &nsec, &dlop.task.tid, (char *)&dlop.sid,
+				     &dlop.base_addr);
+			if (num != 5)
+				goto out;
 
 			pos = strstr(line, "libname=");
 			if (pos == NULL)
-				pr_err_ns("invalid task.txt format");
+				goto out;
+
 			exename = pos + 8 + 1;  // skip double-quote
 			pos = strrchr(exename, '\"');
 			if (pos)
@@ -191,11 +207,17 @@ int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
 					   dlop.base_addr, exename);
 		}
 	}
+	ret = 0;
 
+out:
 	free(line);
 	fclose(fp);
 	free(fname);
-	return 0;
+
+	if (ret != 0)
+		errno = EINVAL;
+
+	return ret;
 }
 
 /**
@@ -207,7 +229,7 @@ int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
  *
  * It returns 0 for success, -1 for error.
  */
-int read_events_file(struct ftrace_file_handle *handle)
+int read_events_file(struct uftrace_data *handle)
 {
 	FILE *fp;
 	char *fname = NULL;
@@ -340,7 +362,7 @@ void write_dlopen_info(const char *dirname, struct uftrace_msg_dlopen *dmsg,
 	free(fname);
 }
 
-static void check_data_order(struct ftrace_file_handle *handle)
+static void check_data_order(struct uftrace_data *handle)
 {
 	union {
 		struct uftrace_record s;
@@ -359,12 +381,44 @@ static void check_data_order(struct ftrace_file_handle *handle)
 		pr_dbg("bitfield order is different!\n");
 }
 
-int open_data_file(struct opts *opts, struct ftrace_file_handle *handle)
+static bool check_data_file(struct uftrace_data *handle,
+			    const char *pattern)
+{
+	glob_t g;
+	size_t i;
+	bool found = false;
+
+	if (glob(pattern, 0, NULL, &g) == GLOB_ERR) {
+		pr_dbg("glob matching failed: %s: %m\n", pattern);
+		return false;
+	}
+
+	for (i = 0; i < g.gl_pathc; i++) {
+		struct stat stbuf;
+
+		if (stat(g.gl_pathv[i], &stbuf) == 0 && stbuf.st_size) {
+			found = true;
+			break;
+		}
+	}
+
+	globfree(&g);
+	return found;
+}
+
+bool data_is_lp64(struct uftrace_data *handle)
+{
+	return handle->hdr.class == ELFCLASS64;
+}
+
+int open_info_file(struct opts *opts, struct uftrace_data *handle)
 {
-	int ret = -1;
 	FILE *fp;
 	char buf[PATH_MAX];
 	int saved_errno = 0;
+	struct stat stbuf;
+
+	memset(handle, 0, sizeof(*handle));
 
 	snprintf(buf, sizeof(buf), "%s/info", opts->dirname);
 
@@ -372,6 +426,11 @@ int open_data_file(struct opts *opts, struct ftrace_file_handle *handle)
 	if (fp != NULL)
 		goto ok;
 
+	saved_errno = errno;
+	/* provide a better error code for empty/invalid directories */
+	if (stat(opts->dirname, &stbuf) == 0)
+		saved_errno = EINVAL;
+
 	/* if default dirname is failed */
 	if (!strcmp(opts->dirname, UFTRACE_DIR_NAME)) {
 		/* try again inside the current directory */
@@ -397,24 +456,18 @@ int open_data_file(struct opts *opts, struct ftrace_file_handle *handle)
 
 	/* data file loading is failed */
 	pr_dbg("cannot open %s file\n", buf);
-	goto out;
 
+	return -saved_errno;
 ok:
+	saved_errno = 0;
 	handle->fp = fp;
 	handle->dirname = opts->dirname;
 	handle->depth = opts->depth;
-	handle->nr_tasks = 0;
-	handle->tasks = NULL;
 	handle->time_filter = opts->threshold;
 	handle->time_range = opts->range;
 	handle->sessions.root  = RB_ROOT;
 	handle->sessions.tasks = RB_ROOT;
-	handle->sessions.first = NULL;
-	handle->kernel = NULL;
-	handle->nr_perf = 0;
-	handle->perf = NULL;
 	handle->last_perf_idx = -1;
-	handle->perf_event_processed = false;
 	INIT_LIST_HEAD(&handle->events);
 
 	if (fread(&handle->hdr, sizeof(handle->hdr), 1, fp) != 1)
@@ -439,37 +492,87 @@ ok:
 	if (read_uftrace_info(handle->hdr.info_mask, handle) < 0)
 		pr_err_ns("cannot read uftrace header info!\n");
 
-	fclose(fp);
-
 	if (opts->exename == NULL)
 		opts->exename = handle->info.exename;
 
+	fclose(fp);
+	return 0;
+}
+
+int open_data_file(struct opts *opts, struct uftrace_data *handle)
+{
+	int ret;
+	char buf[PATH_MAX];
+	int saved_errno = 0;
+
+	ret = open_info_file(opts, handle);
+	if (ret < 0) {
+		errno = -ret;
+		return -1;
+	}
+
+	if (handle->info.nr_tid == 0) {
+		errno = ENODATA;
+		return -1;
+	}
+
 	if (handle->hdr.feat_mask & TASK_SESSION) {
 		bool sym_rel = false;
 		struct uftrace_session_link *sessions = &handle->sessions;
+		int i;
 
 		if (handle->hdr.feat_mask & SYM_REL_ADDR)
 			sym_rel = true;
 
 		/* read old task file first and then try task.txt file */
-		if (read_task_file(sessions, opts->dirname, true, sym_rel) < 0 &&
-		    read_task_txt_file(sessions, opts->dirname, true, sym_rel) < 0) {
+		if (read_task_file(sessions, opts->dirname, true, sym_rel,
+				   opts->srcline) < 0 &&
+		    read_task_txt_file(sessions, opts->dirname, true, sym_rel,
+				       opts->srcline) < 0) {
 			if (errno == ENOENT)
 				saved_errno = ENODATA;
 			else
 				saved_errno = errno;
+
+			goto out;
+		}
+
+		if (sessions->first == NULL) {
+			saved_errno = EINVAL;
+			goto out;
+		}
+
+		for (i = 0; i < handle->info.nr_tid; i++) {
+			int tid = handle->info.tids[i];
+
+			if (find_task(sessions, tid))
+				break;
+		}
+
+		if (i == handle->info.nr_tid) {
+			saved_errno = ENODATA;
+			goto out;
 		}
 	}
 
 	if (handle->hdr.info_mask & ARG_SPEC) {
+		struct uftrace_filter_setting setting = {
+			.ptype		= handle->info.patt_type,
+			.allow_kernel	= true,
+			.auto_args	= false,
+			.lp64		= data_is_lp64(handle),
+			.arch		= handle->arch,
+		};
+
 		if (handle->hdr.feat_mask & AUTO_ARGS) {
 			setup_auto_args_str(handle->info.autoarg,
 					    handle->info.autoret,
-					    handle->info.autoenum);
+					    handle->info.autoenum,
+					    &setting);
 		}
 
 		setup_fstack_args(handle->info.argspec, handle->info.retspec,
-				  handle, false, handle->info.patt_type);
+				  handle, &setting);
 
 		if (handle->info.auto_args_enabled) {
 			char *autoarg = handle->info.autoarg;
@@ -482,8 +585,8 @@ ok:
 					autoarg = autoret = "*";
 			}
 
-			setup_fstack_args(autoarg, autoret,
-					  handle, true, handle->info.patt_type);
+			setting.auto_args = true;
+			setup_fstack_args(autoarg, autoret, handle, &setting);
 		}
 	}
 
@@ -515,16 +618,36 @@ ok:
 	if (handle->hdr.feat_mask & PERF_EVENT)
 		setup_perf_data(handle);
 
-	if (saved_errno)
+	setup_extern_data(handle, opts);
+
+	/* check there are data files actually */
+	snprintf(buf, sizeof(buf), "%s/[0-9]*.dat", opts->dirname);
+	if (!check_data_file(handle, buf)) {
+		if (handle->kernel) {
+			snprintf(buf, sizeof(buf), "%s/kernel-*.dat",
+				 opts->dirname);
+
+			if (check_data_file(handle, buf))
+				goto out;
+		}
+
+		if (saved_errno == 0)
+			saved_errno = ENODATA;
+	}
+
+out:
+	if (saved_errno) {
+		close_data_file(opts, handle);
 		errno = saved_errno;
+		ret = -1;
+	}
 	else
 		ret = 0;
 
-out:
 	return ret;
 }
 
-void close_data_file(struct opts *opts, struct ftrace_file_handle *handle)
+void close_data_file(struct opts *opts, struct uftrace_data *handle)
 {
 	if (opts->exename == handle->info.exename)
 		opts->exename = NULL;
@@ -537,7 +660,11 @@ void close_data_file(struct opts *opts, struct ftrace_file_handle *handle)
 	if (has_perf_data(handle))
 		finish_perf_data(handle);
 
+	if (has_extern_data(handle))
+		finish_extern_data(handle);
+
 	delete_sessions(&handle->sessions);
+	unload_module_symtabs();
 
 	if (handle->hdr.feat_mask & AUTO_ARGS)
 		finish_auto_args();
