@@ -30,6 +30,12 @@ static void snprintf_trigger_read(char *buf, size_t len,
 		snprintf(buf, len, "%s%s", buf[0] ? "|" : "", "proc/statm");
 	if (type & TRIGGER_READ_PAGE_FAULT)
 		snprintf(buf, len, "%s%s", buf[0] ? "|" : "", "page-fault");
+	if (type & TRIGGER_READ_PMU_CYCLE)
+		snprintf(buf, len, "%s%s", buf[0] ? "|" : "", "pmu-cycle");
+	if (type & TRIGGER_READ_PMU_CACHE)
+		snprintf(buf, len, "%s%s", buf[0] ? "|" : "", "pmu-cache");
+	if (type & TRIGGER_READ_PMU_BRANCH)
+		snprintf(buf, len, "%s%s", buf[0] ? "|" : "", "pmu-branch");
 }
 
 static void print_trigger(struct uftrace_trigger *tr)
@@ -82,6 +88,8 @@ static void print_trigger(struct uftrace_trigger *tr)
 		pr_dbg("\ttrigger: color '%c'\n", tr->color);
 	if (tr->flags & TRIGGER_FL_TIME_FILTER)
 		pr_dbg("\ttrigger: time filter %"PRIu64"\n", tr->time);
+	if (tr->flags & TRIGGER_FL_CALLER)
+		pr_dbg("\ttrigger: caller filter\n");
 
 	if (tr->flags & TRIGGER_FL_READ) {
 		char buf[1024];
@@ -91,6 +99,29 @@ static void print_trigger(struct uftrace_trigger *tr)
 	}
 }
 
+/**
+ * uftrace_count_filter - count matching filters in @root
+ * @root - root of rbtree which has filters
+ * @flag - filter flag to match
+ */
+int uftrace_count_filter(struct rb_root *root, unsigned long flag)
+{
+	struct rb_node *entry;
+	struct uftrace_filter *iter;
+	int count = 0;
+
+	entry = rb_first(root);
+	while (entry) {
+		iter = rb_entry(entry, struct uftrace_filter, node);
+
+		if (iter->trigger.flags & flag)
+			count++;
+
+		entry = rb_next(entry);
+	}
+	return count;
+}
+
 static bool match_ip(struct uftrace_filter *filter, unsigned long ip)
 {
 	return filter->start <= ip && ip < filter->end;
@@ -220,8 +251,9 @@ void add_trigger(struct uftrace_filter *filter, struct uftrace_trigger *tr,
 }
 
 static int add_filter(struct rb_root *root, struct uftrace_filter *filter,
-		      struct uftrace_trigger *tr, bool exact_match,
-		      struct debug_info *dinfo)
+		      struct uftrace_trigger *tr, struct uftrace_mmap *map,
+		      bool exact_match, struct debug_info *dinfo,
+		      struct uftrace_filter_setting *setting)
 {
 	struct rb_node *parent = NULL;
 	struct rb_node **p = &root->rb_node;
@@ -231,26 +263,30 @@ static int add_filter(struct rb_root *root, struct uftrace_filter *filter,
 	unsigned long orig_flags = tr->flags;
 
 	if ((tr->flags & TRIGGER_FL_ARGUMENT) && list_empty(tr->pargs)) {
-		auto_arg = find_auto_argspec(filter, tr, dinfo);
+		auto_arg = find_auto_argspec(filter, tr, dinfo, setting);
 		if (auto_arg == NULL)
 			tr->flags &= ~TRIGGER_FL_ARGUMENT;
 	}
 	if ((tr->flags & TRIGGER_FL_RETVAL) && list_empty(tr->pargs)) {
-		auto_ret = find_auto_retspec(filter, tr, dinfo);
+		auto_ret = find_auto_retspec(filter, tr, dinfo, setting);
 		if (auto_ret == NULL)
 			tr->flags &= ~TRIGGER_FL_RETVAL;
 	}
 
-	if (tr->flags == 0 && orig_flags) {
+	/* remove unnecessary filters might be set by --auto-args */
+	if (tr->flags == TRIGGER_FL_AUTO_ARGS) {
 		/* restored for regex filter */
 		tr->flags = orig_flags;
 		return 0;
 	}
 
-	pr_dbg2("add filter for %s\n", filter->name);
+	pr_dbg2("add filter for %s (flags = %x)\n", filter->name, tr->flags);
 	if (dbg_domain[DBG_FILTER] >= 3)
 		print_trigger(tr);
 
+	filter->start += map->start;
+	filter->end   += map->start;
+
 	while (*p) {
 		parent = *p;
 		iter = rb_entry(parent, struct uftrace_filter, node);
@@ -381,18 +417,9 @@ const char * get_filter_pattern(enum uftrace_pattern_type ptype)
 	return "none";
 }
 
-static bool is_arm_machine(void)
+static bool is_arm_machine(struct uftrace_filter_setting *setting)
 {
-	static char *mach = NULL;
-
-	if (mach == NULL) {
-		struct utsname utsbuf;
-
-		uname(&utsbuf);
-		mach = xstrdup(utsbuf.machine);
-	}
-
-	return mach[0] == 'a' && mach[1] == 'r' && mach[2] == 'm';
+	return setting->arch == UFT_CPU_ARM;
 }
 
 static int check_so_cb(struct dl_phdr_info *info, size_t size, void *data)
@@ -420,10 +447,11 @@ static int has_shared_object(const char *soname)
 }
 
 /* argument_spec = arg1/i32,arg2/x64,... */
-static int parse_spec(char *str, struct uftrace_arg_spec *arg, char *suffix)
+static int parse_spec(char *str, struct uftrace_arg_spec *arg, char *suffix,
+		      struct uftrace_filter_setting *setting)
 {
 	int fmt = ARG_FMT_AUTO;
-	int size = sizeof(long);
+	int size = setting->lp64 ? 8 : 4;
 	int type = arg->type;
 	int bit;
 	char *p;
@@ -473,7 +501,7 @@ static int parse_spec(char *str, struct uftrace_arg_spec *arg, char *suffix)
 		fmt = ARG_FMT_STD_STRING;
 		break;
 	case 'p':
-		fmt = ARG_FMT_FUNC_PTR;
+		fmt = ARG_FMT_PTR;
 		break;
 	case 'e':
 		fmt = ARG_FMT_ENUM;
@@ -527,7 +555,7 @@ type:
 			type = ARG_TYPE_STACK;
 		}
 		else {
-			arg->reg_idx = arch_register_index(suffix);
+			arg->reg_idx = arch_register_index(setting->arch, suffix);
 			type = ARG_TYPE_REG;
 
 			if (arg->reg_idx < 0) {
@@ -539,7 +567,7 @@ type:
 
 out:
 	/* it seems ARM falls back 'long double' to 'double' */
-	if (fmt == ARG_FMT_FLOAT && size == 10 && is_arm_machine())
+	if (fmt == ARG_FMT_FLOAT && size == 10 && is_arm_machine(setting))
 		size = 8;
 
 	arg->fmt  = fmt;
@@ -550,7 +578,8 @@ out:
 }
 
 /* argument_spec = arg1/i32,arg2/x64%reg,arg3%stack+1,... */
-static int parse_argument_spec(char *str, struct uftrace_trigger *tr)
+static int parse_argument_spec(char *str, struct uftrace_trigger *tr,
+			       struct uftrace_filter_setting *setting)
 {
 	struct uftrace_arg_spec *arg;
 	char *suffix;
@@ -565,7 +594,7 @@ static int parse_argument_spec(char *str, struct uftrace_trigger *tr)
 	arg->idx = strtol(str+3, &suffix, 0);
 	arg->type = ARG_TYPE_INDEX;
 
-	if (parse_spec(str, arg, suffix) == -1) {
+	if (parse_spec(str, arg, suffix, setting) == -1) {
 		free(arg);
 		return -1;
 	}
@@ -576,7 +605,8 @@ static int parse_argument_spec(char *str, struct uftrace_trigger *tr)
 	return 0;
 }
 /* argument_spec = retval/i32 or retval/x64 ... */
-static int parse_retval_spec(char *str, struct uftrace_trigger *tr)
+static int parse_retval_spec(char *str, struct uftrace_trigger *tr,
+			     struct uftrace_filter_setting *setting)
 {
 	struct uftrace_arg_spec *arg;
 	char *suffix;
@@ -589,7 +619,7 @@ static int parse_retval_spec(char *str, struct uftrace_trigger *tr)
 	/* set suffix after string "retval" */
 	suffix = str + 6;
 
-	if (parse_spec(str, arg, suffix) == -1) {
+	if (parse_spec(str, arg, suffix, setting) == -1) {
 		free(arg);
 		return -1;
 	}
@@ -601,7 +631,8 @@ static int parse_retval_spec(char *str, struct uftrace_trigger *tr)
 }
 
 /* argument_spec = fparg1/32,fparg2/64%stack+1,... */
-static int parse_float_argument_spec(char *str, struct uftrace_trigger *tr)
+static int parse_float_argument_spec(char *str, struct uftrace_trigger *tr,
+				     struct uftrace_filter_setting *setting)
 {
 	struct uftrace_arg_spec *arg;
 	char *suffix;
@@ -626,7 +657,7 @@ static int parse_float_argument_spec(char *str, struct uftrace_trigger *tr)
 			free(arg);
 			return -1;
 		}
-		if (size == 80 && is_arm_machine())
+		if (size == 80 && is_arm_machine(setting))
 			size = 64;
 
 		arg->size = size / 8;
@@ -640,7 +671,7 @@ static int parse_float_argument_spec(char *str, struct uftrace_trigger *tr)
 			arg->type = ARG_TYPE_STACK;
 		}
 		else {
-			arg->reg_idx = arch_register_index(suffix);
+			arg->reg_idx = arch_register_index(setting->arch, suffix);
 			arg->type = ARG_TYPE_REG;
 
 			if (arg->reg_idx < 0) {
@@ -657,7 +688,8 @@ static int parse_float_argument_spec(char *str, struct uftrace_trigger *tr)
 	return 0;
 }
 
-static int parse_depth_action(char *action, struct uftrace_trigger *tr)
+static int parse_depth_action(char *action, struct uftrace_trigger *tr,
+			      struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_DEPTH;
 	tr->depth = strtoul(action + 6, NULL, 10);
@@ -669,14 +701,16 @@ static int parse_depth_action(char *action, struct uftrace_trigger *tr)
 	return 0;
 }
 
-static int parse_time_action(char *action, struct uftrace_trigger *tr)
+static int parse_time_action(char *action, struct uftrace_trigger *tr,
+			     struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_TIME_FILTER;
 	tr->time = parse_time(action + 5, 3);
 	return 0;
 }
 
-static int parse_read_action(char *action, struct uftrace_trigger *tr)
+static int parse_read_action(char *action, struct uftrace_trigger *tr,
+			     struct uftrace_filter_setting *setting)
 {
 	const char *target = action + 5;
 
@@ -698,7 +732,8 @@ static int parse_read_action(char *action, struct uftrace_trigger *tr)
 	return 0;
 }
 
-static int parse_color_action(char *action, struct uftrace_trigger *tr)
+static int parse_color_action(char *action, struct uftrace_trigger *tr,
+			      struct uftrace_filter_setting *setting)
 {
 	const char *color = action + 6;
 
@@ -727,7 +762,8 @@ static int parse_color_action(char *action, struct uftrace_trigger *tr)
 	return 0;
 }
 
-static int parse_trace_action(char *action, struct uftrace_trigger *tr)
+static int parse_trace_action(char *action, struct uftrace_trigger *tr,
+			      struct uftrace_filter_setting *setting)
 {
 	action += 5;
 	if (*action == '_' || *action == '-')
@@ -745,47 +781,61 @@ static int parse_trace_action(char *action, struct uftrace_trigger *tr)
 	return 0;
 }
 
-static int parse_backtrace_action(char *action, struct uftrace_trigger *tr)
+static int parse_backtrace_action(char *action, struct uftrace_trigger *tr,
+				  struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_BACKTRACE;
 	return 0;
 }
 
-static int parse_recover_action(char *action, struct uftrace_trigger *tr)
+static int parse_recover_action(char *action, struct uftrace_trigger *tr,
+				struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_RECOVER;
 	return 0;
 }
 
-static int parse_finish_action(char *action, struct uftrace_trigger *tr)
+static int parse_finish_action(char *action, struct uftrace_trigger *tr,
+			       struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_FINISH;
 	return 0;
 }
 
-static int parse_filter_action(char *action, struct uftrace_trigger *tr)
+static int parse_filter_action(char *action, struct uftrace_trigger *tr,
+			       struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_FILTER;
 	tr->fmode  = FILTER_MODE_IN;
 	return 0;
 }
 
-static int parse_notrace_action(char *action, struct uftrace_trigger *tr)
+static int parse_notrace_action(char *action, struct uftrace_trigger *tr,
+				struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_FILTER;
 	tr->fmode  = FILTER_MODE_OUT;
 	return 0;
 }
 
-static int parse_auto_args_action(char *action, struct uftrace_trigger *tr)
+static int parse_auto_args_action(char *action, struct uftrace_trigger *tr,
+				  struct uftrace_filter_setting *setting)
 {
 	tr->flags |= TRIGGER_FL_ARGUMENT | TRIGGER_FL_RETVAL;
 	return 0;
 }
 
+static int parse_caller_action(char *action, struct uftrace_trigger *tr,
+			       struct uftrace_filter_setting *setting)
+{
+	tr->flags |= TRIGGER_FL_CALLER;
+	return 0;
+}
+
 struct trigger_action_parser {
 	const char *name;
-	int (*parse)(char *action, struct uftrace_trigger *tr);
+	int (*parse)(char *action, struct uftrace_trigger *tr,
+		     struct uftrace_filter_setting *setting);
 	unsigned long flags;
 };
 
@@ -797,17 +847,19 @@ static const struct trigger_action_parser actions[] = {
 	{ "notrace",   parse_notrace_action,      TRIGGER_FL_FILTER, },
 	{ "depth=",    parse_depth_action,        TRIGGER_FL_FILTER, },
 	{ "time=",     parse_time_action,         TRIGGER_FL_FILTER, },
+	{ "caller",    parse_caller_action,       TRIGGER_FL_FILTER, },
+	{ "trace",     parse_trace_action,        TRIGGER_FL_SIGNAL, },
+	{ "finish",    parse_finish_action,       TRIGGER_FL_SIGNAL, },
 	{ "read=",     parse_read_action, },
 	{ "color=",    parse_color_action, },
-	{ "trace",     parse_trace_action, },
 	{ "backtrace", parse_backtrace_action, },
 	{ "recover",   parse_recover_action, },
-	{ "finish",    parse_finish_action, },
 	{ "auto-args", parse_auto_args_action, },
 };
 
 int setup_trigger_action(char *str, struct uftrace_trigger *tr,
-			 char **module, unsigned long orig_flags)
+			 char **module, unsigned long orig_flags,
+			 struct uftrace_filter_setting *setting)
 {
 	char *pos = strchr(str, '@');
 	struct strv acts = STRV_INIT;
@@ -815,6 +867,9 @@ int setup_trigger_action(char *str, struct uftrace_trigger *tr,
 	size_t i;
 	int j;
 
+	if (module != NULL)
+		*module = NULL;
+
 	if (pos == NULL)
 		return 0;
 
@@ -831,14 +886,14 @@ int setup_trigger_action(char *str, struct uftrace_trigger *tr,
 			if (orig_flags && !(orig_flags & action->flags))
 				break;  /* ignore incompatible actions */
 
-			if (action->parse(pos, tr) < 0)
+			if (action->parse(pos, tr, setting) < 0)
 				goto out;
 
 			break;
 		}
 
 		/* if it's not an action, treat it as a module name */
-		if (i == ARRAY_SIZE(actions)) {
+		if (i == ARRAY_SIZE(actions) && module != NULL) {
 			if (*module)
 				pr_use("ignoring extra module: %s\n", pos);
 			else
@@ -848,16 +903,22 @@ int setup_trigger_action(char *str, struct uftrace_trigger *tr,
 	ret = 0;
 
 out:
+	if (ret < 0 && module != NULL)
+		free(*module);
+
 	strv_free(&acts);
 	return ret;
 }
 
-static int add_trigger_entry(struct rb_root *root, struct symtab *symtab,
+static int add_trigger_entry(struct rb_root *root,
 			     struct uftrace_pattern *patt,
 			     struct uftrace_trigger *tr,
-			     struct debug_info *dinfo)
+			     struct uftrace_mmap *map,
+			     struct uftrace_filter_setting *setting)
 {
 	struct uftrace_filter filter;
+	struct symtab *symtab = &map->mod->symtab;
+	struct debug_info *dinfo = &map->mod->dinfo;
 	struct sym *sym;
 	unsigned i;
 	int ret = 0;
@@ -868,12 +929,15 @@ static int add_trigger_entry(struct rb_root *root, struct symtab *symtab,
 		if (!match_filter_pattern(patt, sym->name))
 			continue;
 
+		if (setting->plt_only && sym->type != ST_PLT_FUNC)
+			continue;
+
 		filter.name  = sym->name;
 		filter.start = sym->addr;
 		filter.end   = sym->addr + sym->size;
 
-		ret += add_filter(root, &filter, tr,
-				  patt->type == PATT_SIMPLE, dinfo);
+		ret += add_filter(root, &filter, tr, map,
+				  patt->type == PATT_SIMPLE, dinfo, setting);
 	}
 
 	return ret;
@@ -881,8 +945,8 @@ static int add_trigger_entry(struct rb_root *root, struct symtab *symtab,
 
 static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 			  struct rb_root *root, unsigned long flags,
-			  enum filter_mode *fmode, bool allow_kernel,
-			  enum uftrace_pattern_type ptype)
+			  enum filter_mode *fmode,
+			  struct uftrace_filter_setting *setting)
 {
 	struct strv filters = STRV_INIT;
 	char *name;
@@ -907,11 +971,12 @@ static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 			.type = PATT_NONE,
 		};
 
-		if (setup_trigger_action(name, &tr, &module, flags) < 0)
+		if (setup_trigger_action(name, &tr, &module, flags, setting) < 0)
 			goto next;
 
 		/* skip unintended kernel symbols */
-		if (module && !strcasecmp(module, "kernel") && !allow_kernel)
+		if (module && has_kernel_opt(module) &&
+		    !setting->allow_kernel)
 			goto next;
 
 		if (flags & TRIGGER_FL_FILTER) {
@@ -925,55 +990,43 @@ static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 
 		/* use demangled name for triggers (some auto-args need it) */
 		name = demangle(name);
-		init_filter_pattern(ptype, &patt, name);
+		init_filter_pattern(setting->ptype, &patt, name);
 		free(name);
 
 		if (module) {
-			/* is it the main executable? */
-			if (!strncmp(module, basename(symtabs->filename),
-				     strlen(module))) {
-				ret += add_trigger_entry(root, &symtabs->symtab,
-							 &patt, &tr,
-							 &symtabs->dinfo);
-				ret += add_trigger_entry(root, &symtabs->dsymtab,
-							 &patt, &tr,
-							 &symtabs->dinfo);
+			if (!strcasecmp(module, "PLT")) {
+				setting->plt_only = true;
+				ret += add_trigger_entry(root, &patt, &tr,
+							 symtabs->maps,
+							 setting);
+				setting->plt_only = false;
 			}
-			else if (!strcasecmp(module, "PLT")) {
-				ret = add_trigger_entry(root, &symtabs->dsymtab,
-							&patt, &tr,
-							&symtabs->dinfo);
-			}
-			else if (!strcasecmp(module, "kernel")) {
-				ret = add_trigger_entry(root, get_kernel_symtab(),
-							&patt, &tr,
-							&symtabs->dinfo);
+			else if (has_kernel_opt(module)) {
+				struct uftrace_mmap kernel_map = {
+					.mod = get_kernel_module(),
+				};
+
+				ret = add_trigger_entry(root, &patt, &tr,
+							&kernel_map,
+							setting);
 			}
 			else {
 				map = find_map_by_name(symtabs, module);
-				if (map) {
-					ret = add_trigger_entry(root, &map->symtab,
-								&patt, &tr,
-								&map->dinfo);
+				if (map && map->mod) {
+					ret = add_trigger_entry(root, &patt,
+								&tr, map,
+								setting);
 				}
 			}
-
-			free(module);
 		}
 		else {
-			/* check main executable's symtab first */
-			ret += add_trigger_entry(root, &symtabs->symtab,
-						 &patt, &tr, &symtabs->dinfo);
-			ret += add_trigger_entry(root, &symtabs->dsymtab,
-						 &patt, &tr, &symtabs->dinfo);
+			for_each_map(symtabs, map) {
+				/* some modules don't have symbol table */
+				if (map->mod == NULL)
+					continue;
 
-			/* and then find all module's symtabs */
-			map = symtabs->maps;
-			while (map) {
-				ret += add_trigger_entry(root, &map->symtab,
-							 &patt, &tr,
-							 &map->dinfo);
-				map = map->next;
+				ret += add_trigger_entry(root, &patt, &tr,
+							 map, setting);
 			}
 		}
 
@@ -985,6 +1038,7 @@ static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 		}
 next:
 		free_filter_pattern(&patt);
+		free(module);
 
 		while (!list_empty(&args)) {
 			arg = list_first_entry(&args, typeof(*arg), list);
@@ -1006,16 +1060,14 @@ next:
  * @symtabs    - symbol tables to find symbol address
  * @root       - root of resulting rbtree
  * @mode       - filter mode: opt-in (-F) or opt-out (-N)
- * @allow_kernel - allow filtering on kernel function
- * @patt_type    - filter match pattern (regex or glob)
+ * @setting    - filter settings
  */
 void uftrace_setup_filter(char *filter_str, struct symtabs *symtabs,
 			  struct rb_root *root, enum filter_mode *mode,
-			  bool allow_kernel,
-			  enum uftrace_pattern_type patt_type)
+			  struct uftrace_filter_setting *setting)
 {
 	setup_trigger(filter_str, symtabs, root, TRIGGER_FL_FILTER, mode,
-		      allow_kernel, patt_type);
+		      setting);
 }
 
 /**
@@ -1024,16 +1076,13 @@ void uftrace_setup_filter(char *filter_str, struct symtabs *symtabs,
  * @symtabs    - symbol tables to find symbol address
  * @root       - root of resulting rbtree
  * @mode       - filter mode: opt-in (-F) or opt-out (-N)
- * @allow_kernel - allow filtering on kernel function
- * @patt_type    - filter match pattern (regex or glob)
+ * @setting    - filter settings
  */
 void uftrace_setup_trigger(char *trigger_str, struct symtabs *symtabs,
 			   struct rb_root *root, enum filter_mode *mode,
-			   bool allow_kernel,
-			   enum uftrace_pattern_type patt_type)
+			   struct uftrace_filter_setting *setting)
 {
-	setup_trigger(trigger_str, symtabs, root, 0, mode, allow_kernel,
-		      patt_type);
+	setup_trigger(trigger_str, symtabs, root, 0, mode, setting);
 }
 
 /**
@@ -1041,19 +1090,18 @@ void uftrace_setup_trigger(char *trigger_str, struct symtabs *symtabs,
  * @args_str   - CSV of argument string (FUNC @ arg)
  * @symtabs    - symbol tables to find symbol address
  * @root       - root of resulting rbtree
- * @auto_args  - whether current arguments are auto-spec
- * @patt_type    - filter match pattern (regex or glob)
+ * @setting    - filter settings
  */
 void uftrace_setup_argument(char *args_str, struct symtabs *symtabs,
-			    struct rb_root *root, bool auto_args,
-			    enum uftrace_pattern_type patt_type)
+			    struct rb_root *root,
+			    struct uftrace_filter_setting *setting)
 {
 	unsigned long flags = TRIGGER_FL_ARGUMENT;
 
-	if (auto_args)
+	if (setting->auto_args)
 		flags |= TRIGGER_FL_AUTO_ARGS;
 
-	setup_trigger(args_str, symtabs, root, flags, NULL, false, patt_type);
+	setup_trigger(args_str, symtabs, root, flags, NULL, setting);
 }
 
 /**
@@ -1061,19 +1109,33 @@ void uftrace_setup_argument(char *args_str, struct symtabs *symtabs,
  * @retval_str - CSV of return value string (FUNC @ arg)
  * @symtabs    - symbol tables to find symbol address
  * @root       - root of resulting rbtree
- * @auto_args  - whether current retvals are auto-spec
- * @patt_type    - filter match pattern (regex or glob)
+ * @setting    - filter settings
  */
 void uftrace_setup_retval(char *retval_str, struct symtabs *symtabs,
-			  struct rb_root *root, bool auto_args,
-			  enum uftrace_pattern_type patt_type)
+			  struct rb_root *root,
+			  struct uftrace_filter_setting *setting)
 {
 	unsigned long flags = TRIGGER_FL_RETVAL;
 
-	if (auto_args)
+	if (setting->auto_args)
 		flags |= TRIGGER_FL_AUTO_ARGS;
 
-	setup_trigger(retval_str, symtabs, root, flags, NULL, false, patt_type);
+	setup_trigger(retval_str, symtabs, root, flags, NULL, setting);
+}
+
+/**
+ * uftrace_setup_caller_filter - add caller filters to rbtree
+ * @filter_str - CSV of filter string
+ * @symtabs    - symbol tables to find symbol address
+ * @root       - root of resulting rbtree
+ * @setting    - filter settings
+ */
+void uftrace_setup_caller_filter(char *filter_str, struct symtabs *symtabs,
+				 struct rb_root *root,
+				 struct uftrace_filter_setting *setting)
+{
+	setup_trigger(filter_str, symtabs, root, TRIGGER_FL_CALLER, NULL,
+		      setting);
 }
 
 /**
@@ -1129,13 +1191,13 @@ char * uftrace_clear_kernel(char *filter_str)
 	if (filter_str == NULL)
 		return NULL;
 
-	if (strstr(filter_str, "@kernel") == NULL)
+	if (has_kernel_filter(filter_str) == NULL)
 		return xstrdup(filter_str);
 
 	strv_split(&filters, filter_str, ";");
 
 	strv_for_each(&filters, pos, j) {
-		if (strstr(pos, "@kernel") == NULL)
+		if (has_kernel_filter(pos) == NULL)
 			ret = strjoin(ret, pos, ";");
 	}
 	strv_free(&filters);
@@ -1148,22 +1210,31 @@ char * uftrace_clear_kernel(char *filter_str)
 static void filter_test_load_symtabs(struct symtabs *stabs)
 {
 	static struct sym syms[] = {
-		{ 0x1000, 0x1000, ST_GLOBAL, "foo::foo" },
-		{ 0x2000, 0x1000, ST_GLOBAL, "foo::bar" },
-		{ 0x3000, 0x1000, ST_GLOBAL, "foo::baz1" },
-		{ 0x4000, 0x1000, ST_GLOBAL, "foo::baz2" },
-		{ 0x5000, 0x1000, ST_GLOBAL, "foo::baz3" },
-		{ 0x6000, 0x1000, ST_GLOBAL, "foo::~foo" },
+		{ 0x1000, 0x1000, ST_GLOBAL_FUNC, "foo::foo" },
+		{ 0x2000, 0x1000, ST_GLOBAL_FUNC, "foo::bar" },
+		{ 0x3000, 0x1000, ST_GLOBAL_FUNC, "foo::baz1" },
+		{ 0x4000, 0x1000, ST_GLOBAL_FUNC, "foo::baz2" },
+		{ 0x5000, 0x1000, ST_GLOBAL_FUNC, "foo::baz3" },
+		{ 0x6000, 0x1000, ST_GLOBAL_FUNC, "foo::~foo" },
+		{ 0x21000,0x1000, ST_PLT_FUNC, "malloc" },
+		{ 0x22000,0x1000, ST_PLT_FUNC, "free" },
 	};
-	static struct sym dsyms[] = {
-		{ 0x21000, 0x1000, ST_PLT, "malloc" },
-		{ 0x22000, 0x1000, ST_PLT, "free" },
+	static struct uftrace_module mod = {
+		.symtab = {
+			.sym    = syms,
+			.nr_sym = ARRAY_SIZE(syms),
+		},
+	};
+	static struct uftrace_mmap map = {
+		.mod   = &mod,
+		.start = 0x0,
+		.end   = 0x24000,
 	};
 
-	stabs->symtab.sym = syms;
-	stabs->symtab.nr_sym = ARRAY_SIZE(syms);
-	stabs->dsymtab.sym = dsyms;
-	stabs->dsymtab.nr_sym = ARRAY_SIZE(dsyms);
+	mod.symtab.sym = syms;
+	mod.symtab.nr_sym = ARRAY_SIZE(syms);
+
+	stabs->maps = &map;
 	stabs->loaded = true;
 	stabs->loaded_debug = true;  /* skip DWARF debug info parsing */
 }
@@ -1176,12 +1247,14 @@ TEST_CASE(filter_setup_simple)
 	struct rb_root root = RB_ROOT;
 	struct rb_node *node;
 	struct uftrace_filter *filter;
-	enum uftrace_pattern_type ptype = PATT_SIMPLE;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_SIMPLE,
+	};
 
 	filter_test_load_symtabs(&stabs);
 
 	/* test1: simple method */
-	uftrace_setup_filter("foo::bar", &stabs, &root, NULL, false, ptype);
+	uftrace_setup_filter("foo::bar", &stabs, &root, NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -1194,7 +1267,7 @@ TEST_CASE(filter_setup_simple)
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	/* test2: destructor */
-	uftrace_setup_filter("foo::~foo", &stabs, &root, NULL, false, ptype);
+	uftrace_setup_filter("foo::~foo", &stabs, &root, NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -1207,7 +1280,7 @@ TEST_CASE(filter_setup_simple)
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	/* test3: unknown symbol */
-	uftrace_setup_filter("invalid_name", &stabs, &root, NULL, false, ptype);
+	uftrace_setup_filter("invalid_name", &stabs, &root, NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	return TEST_OK;
@@ -1221,11 +1294,13 @@ TEST_CASE(filter_setup_regex)
 	struct rb_root root = RB_ROOT;
 	struct rb_node *node;
 	struct uftrace_filter *filter;
-	enum uftrace_pattern_type ptype = PATT_REGEX;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_REGEX,
+	};
 
 	filter_test_load_symtabs(&stabs);
 
-	uftrace_setup_filter("^foo::b", &stabs, &root, NULL, false, ptype);
+	uftrace_setup_filter("^foo::b", &stabs, &root, NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -1266,11 +1341,13 @@ TEST_CASE(filter_setup_glob)
 	struct rb_root root = RB_ROOT;
 	struct rb_node *node;
 	struct uftrace_filter *filter;
-	enum uftrace_pattern_type ptype = PATT_GLOB;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_GLOB,
+	};
 
 	filter_test_load_symtabs(&stabs);
 
-	uftrace_setup_filter("foo::b*", &stabs, &root, NULL, false, ptype);
+	uftrace_setup_filter("foo::b*", &stabs, &root, NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -1312,15 +1389,17 @@ TEST_CASE(filter_setup_notrace)
 	struct rb_node *node;
 	struct uftrace_filter *filter;
 	enum filter_mode fmode;
-	enum uftrace_pattern_type ptype = PATT_GLOB;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_GLOB,
+	};
 
 	filter_test_load_symtabs(&stabs);
 
-	uftrace_setup_filter("foo::*", &stabs, &root, &fmode, false, ptype);
+	uftrace_setup_filter("foo::*", &stabs, &root, &fmode, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
-	uftrace_setup_filter("!foo::foo", &stabs, &root, &fmode, false, ptype);
+	uftrace_setup_filter("!foo::foo", &stabs, &root, &fmode, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);  /* overall filter mode doesn't change */
 
@@ -1350,11 +1429,13 @@ TEST_CASE(filter_match)
 	struct rb_root root = RB_ROOT;
 	enum filter_mode fmode;
 	struct uftrace_trigger tr;
-	enum uftrace_pattern_type ptype = PATT_REGEX;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_REGEX,
+	};
 
 	filter_test_load_symtabs(&stabs);
 
-	uftrace_setup_filter("foo::foo", &stabs, &root, &fmode, false, ptype);
+	uftrace_setup_filter("foo::foo", &stabs, &root, &fmode, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
@@ -1389,12 +1470,15 @@ TEST_CASE(trigger_setup_actions)
 	};;
 	struct rb_root root = RB_ROOT;
 	struct uftrace_trigger tr;
-	enum uftrace_pattern_type ptype = PATT_REGEX;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_REGEX,
+		.lp64  = host_is_lp64(),
+	};
 
 	filter_test_load_symtabs(&stabs);
 
 	uftrace_setup_trigger("foo::bar@depth=2", &stabs, &root,
-			      NULL, false, ptype);
+			      NULL, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	memset(&tr, 0, sizeof(tr));
@@ -1403,24 +1487,30 @@ TEST_CASE(trigger_setup_actions)
 	TEST_EQ(tr.depth, 2);
 
 	uftrace_setup_trigger("foo::bar@backtrace", &stabs, &root,
-			      NULL, false, ptype);
+			      NULL, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x2500, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_DEPTH | TRIGGER_FL_BACKTRACE);
 
 	uftrace_setup_trigger("foo::baz1@traceon", &stabs, &root,
-			      NULL, false, ptype);
+			      NULL, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x3000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_TRACE_ON);
 
 	uftrace_setup_trigger("foo::baz3@trace_off,depth=1", &stabs, &root,
-			      NULL, false, ptype);
+			      NULL, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x5000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_TRACE_OFF | TRIGGER_FL_DEPTH);
 	TEST_EQ(tr.depth, 1);
 
+	uftrace_setup_trigger("foo::baz2@caller", &stabs, &root,
+			      NULL, &setting);
+	memset(&tr, 0, sizeof(tr));
+	TEST_NE(uftrace_match_filter(0x4200, &root, &tr), NULL);
+	TEST_EQ(tr.flags, TRIGGER_FL_CALLER);
+
 	uftrace_cleanup_filter(&root);
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
@@ -1435,12 +1525,15 @@ TEST_CASE(trigger_setup_filters)
 	struct rb_root root = RB_ROOT;
 	struct uftrace_trigger tr;
 	enum filter_mode fmode = FILTER_MODE_NONE;
-	enum uftrace_pattern_type ptype = PATT_REGEX;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_REGEX,
+		.lp64  = host_is_lp64(),
+	};
 
 	filter_test_load_symtabs(&stabs);
 
 	uftrace_setup_trigger("foo::bar@depth=2,notrace", &stabs, &root,
-			      &fmode, false, ptype);
+			      &fmode, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_OUT);
 
@@ -1451,7 +1544,7 @@ TEST_CASE(trigger_setup_filters)
 	TEST_EQ(tr.fmode, FILTER_MODE_OUT);
 
 	uftrace_setup_filter("foo::baz1", &stabs, &root,
-			     &fmode, false, ptype);
+			     &fmode, &setting);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
 	memset(&tr, 0, sizeof(tr));
@@ -1460,7 +1553,7 @@ TEST_CASE(trigger_setup_filters)
 	TEST_EQ(tr.fmode, FILTER_MODE_IN);
 
 	uftrace_setup_trigger("foo::baz2@notrace", &stabs, &root,
-			      &fmode, false, ptype);
+			      &fmode, &setting);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
 	memset(&tr, 0, sizeof(tr));
@@ -1468,6 +1561,11 @@ TEST_CASE(trigger_setup_filters)
 	TEST_EQ(tr.flags, TRIGGER_FL_FILTER);
 	TEST_EQ(tr.fmode, FILTER_MODE_OUT);
 
+	uftrace_setup_caller_filter("foo::baz3", &stabs, &root, &setting);
+	memset(&tr, 0, sizeof(tr));
+	TEST_NE(uftrace_match_filter(0x5000, &root, &tr), NULL);
+	TEST_EQ(tr.flags, TRIGGER_FL_CALLER);
+
 	uftrace_cleanup_filter(&root);
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
@@ -1482,13 +1580,16 @@ TEST_CASE(trigger_setup_args)
 	struct rb_root root = RB_ROOT;
 	struct uftrace_trigger tr;
 	struct uftrace_arg_spec *spec;
-	enum uftrace_pattern_type ptype = PATT_REGEX;
+	struct uftrace_filter_setting setting = {
+		.ptype = PATT_REGEX,
+		.lp64  = host_is_lp64(),
+		.arch  = UFT_CPU_X86_64,
+	};
 	int count;
 
 	filter_test_load_symtabs(&stabs);
 
-	uftrace_setup_argument("foo::bar@arg1", &stabs, &root,
-			       false, ptype);
+	uftrace_setup_argument("foo::bar@arg1", &stabs, &root, &setting);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	memset(&tr, 0, sizeof(tr));
@@ -1496,8 +1597,7 @@ TEST_CASE(trigger_setup_args)
 	TEST_EQ(tr.flags, TRIGGER_FL_ARGUMENT);
 	TEST_NE(tr.pargs, NULL);
 
-	uftrace_setup_trigger("foo::bar@arg2/s", &stabs, &root,
-			      NULL, false, ptype);
+	uftrace_setup_trigger("foo::bar@arg2/s", &stabs, &root, NULL, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x2500, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_ARGUMENT);
@@ -1520,7 +1620,7 @@ TEST_CASE(trigger_setup_args)
 	TEST_EQ(count, 2);
 
 	uftrace_setup_argument("foo::baz1@arg1/i32,arg2/x64,fparg1/32,fparg2",
-			       &stabs, &root, false, ptype);
+			       &stabs, &root, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x3999, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_ARGUMENT);
@@ -1560,9 +1660,8 @@ TEST_CASE(trigger_setup_args)
 	}
 	TEST_EQ(count, 4);
 
-	/* FIXME: this test will fail on non-x86 architecture */
 	uftrace_setup_trigger("foo::baz2@arg1/c,arg2/x32%rdi,arg3%stack+4,retval/f64",
-			      &stabs, &root, NULL, false, ptype);
+			      &stabs, &root, NULL, &setting);
 	memset(&tr, 0, sizeof(tr));
 	TEST_NE(uftrace_match_filter(0x4000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_ARGUMENT | TRIGGER_FL_RETVAL);
@@ -1581,8 +1680,7 @@ TEST_CASE(trigger_setup_args)
 			TEST_EQ(spec->fmt, ARG_FMT_HEX);
 			TEST_EQ(spec->type, ARG_TYPE_REG);
 			TEST_EQ(spec->size, 4);
-			/* XXX: x86-specific */
-			TEST_EQ(spec->reg_idx, arch_register_index("rdi"));
+			TEST_EQ(spec->reg_idx, 1);
 			break;
 		case 3:
 			TEST_EQ(spec->idx, 3);
