@@ -5,8 +5,9 @@
 #include <stdint.h>
 #include <regex.h>
 
-#include "rbtree.h"
-#include "list.h"
+#include "utils/rbtree.h"
+#include "utils/list.h"
+#include "utils/arch.h"
 
 /**
  * REGEX_CHARS: characters for regex matching.
@@ -31,6 +32,8 @@ enum trigger_flag {
 	TRIGGER_FL_READ		= (1U << 11),
 	TRIGGER_FL_FINISH	= (1U << 13),
 	TRIGGER_FL_AUTO_ARGS	= (1U << 14),
+	TRIGGER_FL_CALLER	= (1U << 15),
+	TRIGGER_FL_SIGNAL	= (1U << 16),
 };
 
 enum filter_mode {
@@ -48,7 +51,7 @@ enum uftrace_arg_format {
 	ARG_FMT_CHAR,
 	ARG_FMT_FLOAT,
 	ARG_FMT_STD_STRING,
-	ARG_FMT_FUNC_PTR,
+	ARG_FMT_PTR,
 	ARG_FMT_ENUM,
 };
 
@@ -125,6 +128,17 @@ struct uftrace_pattern {
 	regex_t				re;
 };
 
+struct uftrace_filter_setting {
+	enum uftrace_pattern_type	ptype;
+	enum uftrace_cpu_arch		arch;
+	bool				auto_args;
+	bool				allow_kernel;
+	bool				lp64;
+	bool				plt_only;
+	/* caller-defined data */
+	void				*private;
+};
+
 /* please see man proc(5) for /proc/[pid]/statm */
 struct uftrace_proc_statm {
 	uint64_t		vmsize;  /* total program size in KB */
@@ -158,21 +172,25 @@ struct symtabs;
 
 void uftrace_setup_filter(char *filter_str, struct symtabs *symtabs,
 			  struct rb_root *root, enum filter_mode *mode,
-			  bool allow_kernel, enum uftrace_pattern_type ptype);
+			  struct uftrace_filter_setting *setting);
 void uftrace_setup_trigger(char *trigger_str, struct symtabs *symtabs,
 			   struct rb_root *root, enum filter_mode *mode,
-			   bool allow_kernel, enum uftrace_pattern_type ptype);
+			   struct uftrace_filter_setting *setting);
 void uftrace_setup_argument(char *trigger_str, struct symtabs *symtabs,
-			    struct rb_root *root, bool auto_args,
-			    enum uftrace_pattern_type ptype);
+			    struct rb_root *root,
+			    struct uftrace_filter_setting *setting);
 void uftrace_setup_retval(char *trigger_str, struct symtabs *symtabs,
-			  struct rb_root *root, bool auto_args,
-			  enum uftrace_pattern_type ptype);
+			  struct rb_root *root,
+			  struct uftrace_filter_setting *setting);
+void uftrace_setup_caller_filter(char *filter_str, struct symtabs *symtabs,
+				 struct rb_root *root,
+				 struct uftrace_filter_setting *setting);
 
 struct uftrace_filter *uftrace_match_filter(uint64_t ip, struct rb_root *root,
 					    struct uftrace_trigger *tr);
 void uftrace_cleanup_filter(struct rb_root *root);
 void uftrace_print_filter(struct rb_root *root);
+int uftrace_count_filter(struct rb_root *root, unsigned long flag);
 
 void init_filter_pattern(enum uftrace_pattern_type type,
 			 struct uftrace_pattern *p, char *str);
@@ -183,18 +201,21 @@ const char * get_filter_pattern(enum uftrace_pattern_type ptype);
 
 char * uftrace_clear_kernel(char *filter_str);
 
-void setup_auto_args(void);
-void setup_auto_args_str(char *args, char *rets, char *enums);
+void setup_auto_args(struct uftrace_filter_setting *setting);
+void setup_auto_args_str(char *args, char *rets, char *enums,
+			 struct uftrace_filter_setting *setting);
 void finish_auto_args(void);
 
 struct debug_info;
 
 struct uftrace_filter * find_auto_argspec(struct uftrace_filter *filter,
 					  struct uftrace_trigger *tr,
-					  struct debug_info *dinfo);
+					  struct debug_info *dinfo,
+					  struct uftrace_filter_setting *setting);
 struct uftrace_filter * find_auto_retspec(struct uftrace_filter *filter,
 					  struct uftrace_trigger *tr,
-					  struct debug_info *dinfo);
+					  struct debug_info *dinfo,
+					  struct uftrace_filter_setting *setting);
 char *get_auto_argspec_str(void);
 char *get_auto_retspec_str(void);
 char *get_auto_enum_str(void);
@@ -206,4 +227,10 @@ void release_enum_def(struct rb_root *root);
 
 extern struct rb_root dwarf_enum;
 
+void add_trigger(struct uftrace_filter *filter, struct uftrace_trigger *tr,
+		 bool exact_match);
+int setup_trigger_action(char *str, struct uftrace_trigger *tr, char **module,
+			 unsigned long orig_flags,
+			 struct uftrace_filter_setting *setting);
+
 #endif /* UFTRACE_FILTER_H */
