@@ -31,21 +31,23 @@ struct time_filter_stack {
 	enum context context;
 };
 
-struct ftrace_task_handle {
+struct uftrace_task_reader {
 	int tid;
 	bool valid;
 	bool done;
 	bool lost_seen;
+	bool sched_out_seen;
 	bool fork_handled;
 	bool fstack_set;
 	bool display_depth_set;
 	FILE *fp;
 	struct sym *func;
 	struct uftrace_task *t;
-	struct ftrace_file_handle *h;
+	struct uftrace_data *h;
 	struct uftrace_record ustack;
 	struct uftrace_record kstack;
 	struct uftrace_record estack;
+	struct uftrace_record xstack;
 	struct uftrace_record *rstack;
 	struct uftrace_rstack_list rstack_list;
 	struct uftrace_rstack_list event_list;
@@ -57,6 +59,7 @@ struct ftrace_task_handle {
 	int fork_display_depth;
 	int column_index;
 	int event_color;
+	int sched_cpu;
 	enum context ctx;
 	uint64_t timestamp;
 	uint64_t timestamp_last;
@@ -98,60 +101,87 @@ enum argspec_string_bits {
 extern bool fstack_enabled;
 extern bool live_disabled;
 
-struct ftrace_task_handle *get_task_handle(struct ftrace_file_handle *handle,
+struct uftrace_task_reader *get_task_handle(struct uftrace_data *handle,
 					   int tid);
-void reset_task_handle(struct ftrace_file_handle *handle);
+void reset_task_handle(struct uftrace_data *handle);
 
-int read_rstack(struct ftrace_file_handle *handle,
-		struct ftrace_task_handle **task);
-int peek_rstack(struct ftrace_file_handle *handle,
-		struct ftrace_task_handle **task);
-void fstack_consume(struct ftrace_file_handle *handle,
-		    struct ftrace_task_handle *task);
+int read_rstack(struct uftrace_data *handle,
+		struct uftrace_task_reader **task);
+int peek_rstack(struct uftrace_data *handle,
+		struct uftrace_task_reader **task);
+void fstack_consume(struct uftrace_data *handle,
+		    struct uftrace_task_reader *task);
 
-int read_task_ustack(struct ftrace_file_handle *handle,
-		     struct ftrace_task_handle *task);
-int read_task_args(struct ftrace_task_handle *task,
+int read_task_ustack(struct uftrace_data *handle,
+		     struct uftrace_task_reader *task);
+int read_task_args(struct uftrace_task_reader *task,
 		   struct uftrace_record *rstack,
 		   bool is_retval);
 
-static inline bool is_user_record(struct ftrace_task_handle *task,
+static inline bool is_user_record(struct uftrace_task_reader *task,
 				  struct uftrace_record *rec)
 {
 	return rec == &task->ustack;
 }
 
-static inline bool is_kernel_record(struct ftrace_task_handle *task,
+static inline bool is_kernel_record(struct uftrace_task_reader *task,
 				    struct uftrace_record *rec)
 {
 	return rec == &task->kstack;
 }
 
-static inline bool is_event_record(struct ftrace_task_handle *task,
+static inline bool is_event_record(struct uftrace_task_reader *task,
 				  struct uftrace_record *rec)
 {
 	return rec == &task->estack;
 }
 
-void setup_fstack_args(char *argspec, char *retspec,
-		       struct ftrace_file_handle *handle, bool auto_args,
-		       enum uftrace_pattern_type patt_type);
-int fstack_setup_filters(struct opts *opts, struct ftrace_file_handle *handle);
+static inline bool is_extern_record(struct uftrace_task_reader *task,
+				    struct uftrace_record *rec)
+{
+	return rec == &task->xstack;
+}
 
-int fstack_entry(struct ftrace_task_handle *task,
+void setup_fstack_args(char *argspec, char *retspec,
+		       struct uftrace_data *handle,
+		       struct uftrace_filter_setting *setting);
+int fstack_setup_filters(struct opts *opts, struct uftrace_data *handle);
+
+int fstack_entry(struct uftrace_task_reader *task,
 		 struct uftrace_record *rstack,
 		 struct uftrace_trigger *tr);
-void fstack_exit(struct ftrace_task_handle *task);
-int fstack_update(int type, struct ftrace_task_handle *task,
+void fstack_exit(struct uftrace_task_reader *task);
+int fstack_update(int type, struct uftrace_task_reader *task,
 		  struct fstack *fstack);
-struct ftrace_task_handle *fstack_skip(struct ftrace_file_handle *handle,
-				       struct ftrace_task_handle *task,
-				       int curr_depth, bool event_skip_out);
-bool fstack_check_filter(struct ftrace_task_handle *task);
-bool fstack_check_opts(struct ftrace_task_handle *task, struct opts *opts);
+struct uftrace_task_reader *fstack_skip(struct uftrace_data *handle,
+				       struct uftrace_task_reader *task,
+				       int curr_depth, struct opts *opts);
+bool fstack_check_filter(struct uftrace_task_reader *task);
+bool fstack_check_opts(struct uftrace_task_reader *task, struct opts *opts);
 
-void get_argspec_string(struct ftrace_task_handle *task,
+void get_argspec_string(struct uftrace_task_reader *task,
 		        char *args, size_t len,
 		        enum argspec_string_bits str_mode);
 
+#define EXTERN_DATA_MAX  1024
+
+struct uftrace_extern_reader {
+	FILE			*fp;
+	bool			valid;
+	uint64_t		time;
+	char			msg[EXTERN_DATA_MAX];
+	struct uftrace_record	rec;
+};
+
+int setup_extern_data(struct uftrace_data *handle, struct opts *opts);
+int read_extern_data(struct uftrace_extern_reader *extn);
+struct uftrace_record * get_extern_record(struct uftrace_extern_reader *extn,
+					  struct uftrace_record *rec);
+int finish_extern_data(struct uftrace_data *handle);
+
+static inline bool has_extern_data(struct uftrace_data *handle)
+{
+	return handle->extn != NULL;
+}
+
 #endif /* UFTRACE_FSTACK_H */
