@@ -8,12 +8,17 @@
 #ifndef UFTRACE_SCRIPT_PYTHON_H
 #define UFTRACE_SCRIPT_PYTHON_H
 
+#include "utils/filter.h"
+
+struct script_info;
+
 #ifdef HAVE_LIBPYTHON2
 
 #include <python2.7/Python.h>
 
 #define SCRIPT_ENABLED 1
-int script_init_for_python(char *py_pathname);
+int script_init_for_python(struct script_info *info,
+			   enum uftrace_pattern_type ptype);
 void script_finish_for_python(void);
 
 
@@ -22,7 +27,8 @@ void script_finish_for_python(void);
 
 /* Do nothing if libpython2.7.so is not installed. */
 #define SCRIPT_ENABLED 0
-static inline int script_init_for_python(char *py_pathname)
+static inline int script_init_for_python(struct script_info *info,
+					 enum uftrace_pattern_type ptype)
 {
 	return -1;
 }
